import io

from cog import BasePredictor, File
from pydantic import BaseModel


class Output(BaseModel):
    text: str
    file: File


class Predictor(BasePredictor):
    def predict(self) -> Output:
        return Output(text="hello", file=io.StringIO("hello"))



import io
import pathlib
from datetime import datetime
from enum import Enum
from types import GeneratorType
from typing import Any, Callable

from pydantic import BaseModel

# numpy is an optional dependency, but the process of importing it is not
# thread-safe, so we attempt the import once here.
try:
    import numpy as np  # type: ignore
except ImportError:
    np = None

from .types import PYDANTIC_V2


def make_encodeable(obj: Any) -> Any:  # pylint: disable=too-many-return-statements
    """
    Returns a pickle-compatible version of the object. It will encode any Pydantic models and custom types.

    It is almost JSON-compatible. Files must be done in a separate step with upload_files().

    Somewhat based on FastAPI's jsonable_encoder().
    """

    if isinstance(obj, BaseModel):
        if PYDANTIC_V2:
            return make_encodeable(obj.model_dump(exclude_unset=True))
        else:
            return make_encodeable(obj.dict())
    if isinstance(obj, dict):
        return {key: make_encodeable(value) for key, value in obj.items()}
    if isinstance(obj, (list, set, frozenset, GeneratorType, tuple)):
        return [make_encodeable(value) for value in obj]
    if isinstance(obj, Enum):
        return obj.value
    if isinstance(obj, datetime):
        return obj.isoformat()
    if np:
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
    return obj


def upload_files(obj: Any, upload_file: Callable[[io.IOBase], str]) -> Any:
    """
    Iterates through an object from make_encodeable and uploads any files.

    When a file is encountered, it will be passed to upload_file. Any paths will be opened and converted to files.
    """
    # skip four isinstance checks for fast text models
    if type(obj) == str:  # noqa: E721 # pylint: disable=unidiomatic-typecheck
        return obj
    if isinstance(obj, dict):
        return {key: upload_files(value, upload_file) for key, value in obj.items()}
    if isinstance(obj, list):
        return [upload_files(value, upload_file) for value in obj]
    if isinstance(obj, pathlib.Path):
        with obj.open("rb") as f:
            return upload_file(f)
    if isinstance(obj, io.IOBase):
        return upload_file(obj)
    return obj



# Prediction interface for Cog ⚙️
# https://cog.run/python

from cog import BasePredictor, Input, Path


class Predictor(BasePredictor):
    def setup(self) -> None:
        """Load the model into memory to make running multiple predictions efficient"""
        # self.model = torch.load("./weights.pth")

    def predict(
        self,
        image: Path = Input(description="Grayscale input image"),
        scale: float = Input(
            description="Factor to scale image by", ge=0, le=10, default=1.5
        ),
    ) -> Path:
        """Run a single prediction on the model"""
        # processed_input = preprocess(image)
        # output = self.model(processed_image, scale)
        # return postprocess(output)



import requests
import responses
from responses import registries

from cog.schema import PredictionResponse, Status, WebhookEvent
from cog.server.webhook import webhook_caller, webhook_caller_filtered


@responses.activate
def test_webhook_caller_basic():
    c = webhook_caller("https://example.com/webhook/123")

    payload = {
        "status": Status.PROCESSING,
        "output": {"animal": "giraffe"},
        "input": {},
    }
    response = PredictionResponse(**payload)

    responses.post(
        "https://example.com/webhook/123",
        json=payload,
        status=200,
    )

    c(response)


@responses.activate
def test_webhook_caller_non_terminal_does_not_retry():
    c = webhook_caller("https://example.com/webhook/123")

    payload = {
        "status": Status.PROCESSING,
        "output": {"animal": "giraffe"},
        "input": {},
    }
    response = PredictionResponse(**payload)

    responses.post(
        "https://example.com/webhook/123",
        json=payload,
        status=429,
    )

    c(response)


@responses.activate(registry=registries.OrderedRegistry)
def test_webhook_caller_terminal_retries():
    c = webhook_caller("https://example.com/webhook/123")
    resps = []

    payload = {"status": Status.SUCCEEDED, "output": {"animal": "giraffe"}, "input": {}}
    response = PredictionResponse(**payload)

    for _ in range(2):
        resps.append(
            responses.post(
                "https://example.com/webhook/123",
                json=payload,
                status=429,
            )
        )
    resps.append(
        responses.post(
            "https://example.com/webhook/123",
            json=payload,
            status=200,
        )
    )

    c(response)

    assert all(r.call_count == 1 for r in resps)


@responses.activate
def test_webhook_includes_user_agent():
    c = webhook_caller("https://example.com/webhook/123")

    payload = {
        "status": Status.PROCESSING,
        "output": {"animal": "giraffe"},
        "input": {},
    }
    response = PredictionResponse(**payload)

    responses.post(
        "https://example.com/webhook/123",
        json=payload,
        status=200,
    )

    c(response)

    assert len(responses.calls) == 1
    user_agent = responses.calls[0].request.headers["user-agent"]
    assert user_agent.startswith("cog-worker/")


@responses.activate
def test_webhook_caller_filtered_basic():
    events = WebhookEvent.default_events()
    c = webhook_caller_filtered("https://example.com/webhook/123", events)

    payload = {"status": Status.PROCESSING, "animal": "giraffe", "input": {}}
    response = PredictionResponse(**payload)

    responses.post(
        "https://example.com/webhook/123",
        json=payload,
        status=200,
    )

    c(response, WebhookEvent.LOGS)


@responses.activate
def test_webhook_caller_filtered_omits_filtered_events():
    events = {WebhookEvent.COMPLETED}
    c = webhook_caller_filtered("https://example.com/webhook/123", events)

    payload = {
        "status": Status.PROCESSING,
        "output": {"animal": "giraffe"},
        "input": {},
    }
    response = PredictionResponse(**payload)

    c(response, WebhookEvent.LOGS)


@responses.activate
def test_webhook_caller_connection_errors():
    connerror_resp = responses.Response(
        responses.POST,
        "https://example.com/webhook/123",
        status=200,
    )
    connerror_exc = requests.ConnectionError("failed to connect")
    connerror_exc.response = connerror_resp
    connerror_resp.body = connerror_exc
    responses.add(connerror_resp)

    payload = {
        "status": Status.PROCESSING,
        "output": {"animal": "giraffe"},
        "input": {},
    }
    response = PredictionResponse(**payload)

    c = webhook_caller("https://example.com/webhook/123")
    # this should not raise an error
    c(response)



import io
from unittest.mock import Mock

import requests

from cog.files import put_file_to_signed_endpoint


def test_put_file_to_signed_endpoint():
    mock_fh = io.BytesIO()
    mock_client = Mock()

    mock_response = Mock(spec=requests.Response)
    mock_response.status_code = 201
    mock_response.text = ""
    mock_response.headers = {}
    mock_response.url = "http://example.com/upload/file?some-gubbins"
    mock_response.ok = True

    mock_client.put.return_value = mock_response

    final_url = put_file_to_signed_endpoint(
        mock_fh, "http://example.com/upload", mock_client, prediction_id=None
    )

    assert final_url == "http://example.com/upload/file"
    mock_client.put.assert_called_with(
        "http://example.com/upload/file",
        mock_fh,
        headers={
            "Content-Type": None,
        },
        timeout=(10, 15),
    )


def test_put_file_to_signed_endpoint_with_prediction_id():
    mock_fh = io.BytesIO()
    mock_client = Mock()

    mock_response = Mock(spec=requests.Response)
    mock_response.status_code = 201
    mock_response.text = ""
    mock_response.headers = {}
    mock_response.url = "http://example.com/upload/file?some-gubbins"
    mock_response.ok = True

    mock_client.put.return_value = mock_response

    final_url = put_file_to_signed_endpoint(
        mock_fh, "http://example.com/upload", mock_client, prediction_id="abc123"
    )

    assert final_url == "http://example.com/upload/file"
    mock_client.put.assert_called_with(
        "http://example.com/upload/file",
        mock_fh,
        headers={
            "Content-Type": None,
            "X-Prediction-ID": "abc123",
        },
        timeout=(10, 15),
    )


def test_put_file_to_signed_endpoint_with_location():
    mock_fh = io.BytesIO()
    mock_client = Mock()

    mock_response = Mock(spec=requests.Response)
    mock_response.status_code = 201
    mock_response.text = ""
    mock_response.headers = {
        "location": "http://cdn.example.com/bucket/file?some-gubbins"
    }
    mock_response.url = "http://example.com/upload/file?some-gubbins"
    mock_response.ok = True

    mock_client.put.return_value = mock_response

    final_url = put_file_to_signed_endpoint(
        mock_fh, "http://example.com/upload", mock_client, prediction_id="abc123"
    )

    assert final_url == "http://cdn.example.com/bucket/file"
    mock_client.put.assert_called_with(
        "http://example.com/upload/file",
        mock_fh,
        headers={
            "Content-Type": None,
            "X-Prediction-ID": "abc123",
        },
        timeout=(10, 15),
    )



import abc
import asyncio
import collections.abc
import multiprocessing
from multiprocessing.connection import Connection
from typing import Any, Optional

# Buffer is only available in typing-extensions>=4.6.0 but should be available in stdlib
# python 3.12+. This compatibility code is nearly identical to the implementation in
# typing-extensions>=4.6.0
if hasattr(collections.abc, "Buffer"):
    Buffer = collections.abc.Buffer  # type: ignore
else:

    class Buffer(abc.ABC):  # noqa: B024
        pass

    Buffer.register(memoryview)
    Buffer.register(bytearray)
    Buffer.register(bytes)

_spawn = multiprocessing.get_context("spawn")


class AsyncConnection:
    def __init__(self, connection: Connection) -> None:
        self._connection = connection
        self._event = asyncio.Event()
        loop = asyncio.get_event_loop()
        loop.add_reader(self._connection.fileno(), self._event.set)

    def send(self, obj: Any) -> None:
        """Send a (picklable) object"""

        self._connection.send(obj)

    async def _wait_for_input(self) -> None:
        """Wait until there is an input available to be read"""

        while not self._connection.poll():
            await self._event.wait()
            self._event.clear()

    async def recv(self) -> Any:
        """Receive a (picklable) object"""

        await self._wait_for_input()
        return self._connection.recv()

    def fileno(self) -> int:
        """File descriptor or handle of the connection"""
        return self._connection.fileno()

    def close(self) -> None:
        """Close the connection"""
        self._connection.close()

    async def poll(self, timeout: float = 0.0) -> bool:
        """Whether there is an input available to be read"""

        if self._connection.poll():
            return True

        try:
            await asyncio.wait_for(self._wait_for_input(), timeout=timeout)
        except asyncio.TimeoutError:
            return False
        return self._connection.poll()

    def send_bytes(
        self, buf: Buffer, offset: int = 0, size: Optional[int] = None
    ) -> None:
        """Send the bytes data from a bytes-like object"""

        self._connection.send_bytes(buf, offset, size)  # type: ignore

    async def recv_bytes(self, maxlength: Optional[int] = None) -> bytes:
        """
        Receive bytes data as a bytes object.
        """

        await self._wait_for_input()
        return self._connection.recv_bytes(maxlength)

    async def recv_bytes_into(self, buf: Buffer, offset: int = 0) -> int:
        """
        Receive bytes data into a writeable bytes-like object.
        Return the number of bytes read.
        """

        await self._wait_for_input()
        return self._connection.recv_bytes_into(buf, offset)


class LockedConnection:
    def __init__(self, connection: Connection) -> None:
        self.connection = connection
        self._lock = _spawn.Lock()

    def send(self, obj: Any) -> None:
        with self._lock:
            self.connection.send(obj)

    def recv(self) -> Any:
        return self.connection.recv()



import asyncio

from cog import BasePredictor


class Predictor(BasePredictor):
    async def predict(self, s: str, sleep: float) -> str:
        await asyncio.sleep(sleep)
        return f"wake up {s}"



class CogError(Exception):
    """Base class for all Cog errors."""


class ConfigDoesNotExist(CogError):
    """Exception raised when a cog.yaml does not exist."""


class PredictorNotSet(CogError):
    """Exception raised when 'predict' is not set in cog.yaml when it needs to be."""



import inspect
import os
import sys
import uuid
from typing import Any, Callable, Optional, Tuple, Type

import structlog
import yaml
from pydantic import BaseModel

from .base_input import BaseInput
from .base_predictor import BasePredictor
from .code_xforms import load_module_from_string, strip_model_source_code
from .env_property import env_property
from .errors import ConfigDoesNotExist
from .mode import Mode
from .predictor import (
    get_input_type,
    get_output_type,
    get_predict,
    get_predictor,
    get_train,
    get_training_input_type,
    get_training_output_type,
    load_full_predictor_from_file,
)
from .types import CogConfig
from .wait import wait_for_env

COG_YAML_FILE = "cog.yaml"
COG_PREDICT_TYPE_STUB_ENV_VAR = "COG_PREDICT_TYPE_STUB"
COG_TRAIN_TYPE_STUB_ENV_VAR = "COG_TRAIN_TYPE_STUB"
COG_PREDICT_CODE_STRIP_ENV_VAR = "COG_PREDICT_CODE_STRIP"
COG_TRAIN_CODE_STRIP_ENV_VAR = "COG_TRAIN_CODE_STRIP"
COG_GPU_ENV_VAR = "COG_GPU"
COG_MAX_CONCURRENCY_ENV_VAR = "COG_MAX_CONCURRENCY"
PREDICT_METHOD_NAME = "predict"
TRAIN_METHOD_NAME = "train"

log = structlog.get_logger("cog.config")


def _method_name_from_mode(mode: Mode) -> str:
    if mode == Mode.PREDICT:
        return PREDICT_METHOD_NAME
    elif mode == Mode.TRAIN:
        return TRAIN_METHOD_NAME
    raise ValueError(f"Mode {mode} not recognised for method name mapping")


def _env_var_from_mode(mode: Mode) -> str:
    if mode == Mode.PREDICT:
        return COG_PREDICT_CODE_STRIP_ENV_VAR
    elif mode == Mode.TRAIN:
        return COG_TRAIN_CODE_STRIP_ENV_VAR
    raise ValueError(f"Mode {mode} not recognised for env var mapping")


class Config:
    """A class for reading the cog.yaml properties."""

    def __init__(self, config: Optional[CogConfig] = None) -> None:
        self._config = config

    @property
    def _cog_config(self) -> CogConfig:
        """
        Warning: Do not access this directly outside this class, instead
        write an explicit public property and back it by an @env_property
        to allow for the possibility of injecting the property you are
        trying to read without relying on the underlying file.
        """
        config = self._config
        if config is None:
            wait_for_env(include_imports=False)
            config_path = os.path.abspath(COG_YAML_FILE)
            try:
                with open(config_path, encoding="utf-8") as handle:
                    config = yaml.safe_load(handle)
            except FileNotFoundError as e:
                raise ConfigDoesNotExist(
                    f"Could not find {config_path}",
                ) from e
            self._config = config
        return config

    @property
    @env_property(COG_PREDICT_TYPE_STUB_ENV_VAR)
    def predictor_predict_ref(self) -> Optional[str]:
        """Find the predictor ref for the predict mode."""
        return self._cog_config.get(str(Mode.PREDICT))

    @property
    @env_property(COG_TRAIN_TYPE_STUB_ENV_VAR)
    def predictor_train_ref(self) -> Optional[str]:
        """Find the predictor ref for the train mode."""
        return self._cog_config.get(str(Mode.TRAIN))

    @property
    @env_property(COG_GPU_ENV_VAR)
    def requires_gpu(self) -> bool:
        """Whether this cog requires the use of a GPU."""
        return bool(self._cog_config.get("build", {}).get("gpu", False))

    @property
    @env_property(COG_MAX_CONCURRENCY_ENV_VAR)
    def max_concurrency(self) -> int:
        """The maximum concurrency of predictions supported by this model. Defaults to 1."""
        return int(self._cog_config.get("concurrency", {}).get("max", 1))

    def _predictor_code(
        self,
        module_path: str,
        class_name: str,
        method_name: str,
        mode: Mode,
        module_name: str,
    ) -> Optional[str]:
        source_code = os.environ.get(_env_var_from_mode(mode))
        if source_code is not None:
            return source_code
        if sys.version_info >= (3, 9):
            wait_for_env(include_imports=False)
            with open(module_path, encoding="utf-8") as file:
                return strip_model_source_code(file.read(), [class_name], [method_name])
        else:
            log.debug(f"[{module_name}] cannot use fast loader as current Python <3.9")
        return None

    def _load_predictor_for_types(
        self, ref: str, method_name: str, mode: Mode
    ) -> BasePredictor:
        module_path, class_name = ref.split(":", 1)
        module_name = os.path.basename(module_path).split(".py", 1)[0]
        code = self._predictor_code(
            module_path, class_name, method_name, mode, module_name
        )
        module = None
        if code is not None:
            try:
                module = load_module_from_string(uuid.uuid4().hex, code)
            except Exception as e:  # pylint: disable=broad-exception-caught
                log.info(f"[{module_name}] fast loader failed: {e}")
        if module is None:
            log.debug(f"[{module_name}] falling back to slow loader")
            wait_for_env(include_imports=False)
            module = load_full_predictor_from_file(module_path, module_name)
        return get_predictor(module, class_name)

    def get_predictor_ref(self, mode: Mode) -> str:
        """Find the predictor reference for a given mode."""
        predictor_ref = None
        if mode == Mode.PREDICT:
            predictor_ref = self.predictor_predict_ref
        elif mode == Mode.TRAIN:
            predictor_ref = self.predictor_train_ref
        if predictor_ref is None:
            raise ValueError(
                f"Can't run predictions: '{mode}' option not found in cog.yaml"
            )
        return predictor_ref

    def get_predictor_types(
        self, mode: Mode
    ) -> Tuple[Type[BaseInput], Type[BaseModel], bool]:
        """
        Find the input & output types of a predictor/train function as well
        as determining if the function is an async function.
        """
        predictor_ref = self.get_predictor_ref(mode=mode)
        predictor = self._load_predictor_for_types(
            predictor_ref, _method_name_from_mode(mode=mode), mode
        )

        def is_async(fn: Callable[[Any], Any]) -> bool:
            return inspect.iscoroutinefunction(fn) or inspect.isasyncgenfunction(fn)

        if mode == Mode.PREDICT:
            return (
                get_input_type(predictor),
                get_output_type(predictor),
                is_async(get_predict(predictor)),
            )
        elif mode == Mode.TRAIN:
            return (
                get_training_input_type(predictor),
                get_training_output_type(predictor),
                is_async(get_train(predictor)),
            )
        raise ValueError(f"Mode {mode} not found for generating input/output types.")



import os
import tempfile

from cog import BasePredictor, Path


class Predictor(BasePredictor):
    def predict(self) -> Path:
        temp_dir = tempfile.mkdtemp()
        temp_path = os.path.join(temp_dir, "file.txt")
        with open(temp_path, "w") as fh:
            fh.write("hello")
        return Path(temp_path)



import logging
import os
import tempfile
from unittest import mock

import pytest

from cog.server.probes import ProbeHelper


@pytest.fixture
def tmpdir():
    with tempfile.TemporaryDirectory() as td:
        yield td


@mock.patch.dict(os.environ, {"KUBERNETES_SERVICE_HOST": "0.0.0.0"})
def test_ready(tmpdir):
    p = ProbeHelper(root=tmpdir)

    p.ready()

    assert os.path.isfile(os.path.join(tmpdir, "ready"))


def test_does_nothing_when_not_in_k8s(tmpdir, caplog):
    with caplog.at_level(logging.INFO):
        p = ProbeHelper(root=tmpdir)
        p.ready()

    assert os.listdir(tmpdir) == []
    assert "disabling probe helpers" in caplog.text


@mock.patch.dict(os.environ, {"KUBERNETES_SERVICE_HOST": "0.0.0.0"})
def test_creates_probe_dir_if_needed(tmpdir):
    root = os.path.join(tmpdir, "probes")
    p = ProbeHelper(root=root)

    p.ready()

    assert os.path.isdir(os.path.join(tmpdir, "probes"))
    assert os.path.isfile(os.path.join(tmpdir, "probes", "ready"))


@mock.patch.dict(os.environ, {"KUBERNETES_SERVICE_HOST": "0.0.0.0"})
def test_no_exception_when_probe_dir_exists(tmpdir, caplog):
    root = os.path.join(tmpdir, "probes")

    # Create a file
    open(root, "a").close()

    p = ProbeHelper(root=root)
    p.ready()

    assert "Failed to create cog runtime state directory" in caplog.text



import os
import tempfile

import pytest

from cog.config import (
    COG_GPU_ENV_VAR,
    COG_PREDICT_CODE_STRIP_ENV_VAR,
    COG_PREDICT_TYPE_STUB_ENV_VAR,
    COG_TRAIN_TYPE_STUB_ENV_VAR,
    COG_YAML_FILE,
    Config,
)
from cog.errors import ConfigDoesNotExist
from cog.mode import Mode


def test_predictor_predict_ref_env_var():
    predict_ref = "predict.py:Predictor"
    os.environ[COG_PREDICT_TYPE_STUB_ENV_VAR] = predict_ref
    config = Config()
    config_predict_ref = config.predictor_predict_ref
    del os.environ[COG_PREDICT_TYPE_STUB_ENV_VAR]
    assert config_predict_ref == predict_ref, (
        "Predict Reference should come from the environment variable."
    )


def test_predictor_predict_ref_no_env_var():
    if COG_PREDICT_TYPE_STUB_ENV_VAR in os.environ:
        del os.environ[COG_PREDICT_TYPE_STUB_ENV_VAR]
    pwd = os.getcwd()
    with tempfile.TemporaryDirectory() as tmpdir:
        os.chdir(tmpdir)
        with open(COG_YAML_FILE, "w", encoding="utf-8") as handle:
            handle.write("""
build:
  python_version: "3.11"
predict: "predict.py:Predictor"
""")
        config = Config()
        config_predict_ref = config.predictor_predict_ref
        assert config_predict_ref == "predict.py:Predictor", (
            "Predict Reference should come from the cog config file."
        )
    os.chdir(pwd)


def test_config_no_config_file():
    if COG_PREDICT_TYPE_STUB_ENV_VAR in os.environ:
        del os.environ[COG_PREDICT_TYPE_STUB_ENV_VAR]
    config = Config()
    with pytest.raises(ConfigDoesNotExist):
        _ = config.predictor_predict_ref


def test_config_initial_values():
    if COG_PREDICT_TYPE_STUB_ENV_VAR in os.environ:
        del os.environ[COG_PREDICT_TYPE_STUB_ENV_VAR]
    config = Config(config={"predict": "predict.py:Predictor"})
    config_predict_ref = config.predictor_predict_ref
    assert config_predict_ref == "predict.py:Predictor", (
        "Predict Reference should come from the initial config dictionary."
    )


def test_predictor_train_ref_env_var():
    train_ref = "predict.py:Predictor"
    os.environ[COG_TRAIN_TYPE_STUB_ENV_VAR] = train_ref
    config = Config()
    config_train_ref = config.predictor_train_ref
    del os.environ[COG_TRAIN_TYPE_STUB_ENV_VAR]
    assert config_train_ref == train_ref, (
        "Train Reference should come from the environment variable."
    )


def test_predictor_train_ref_no_env_var():
    train_ref = "predict.py:Predictor"
    if COG_TRAIN_TYPE_STUB_ENV_VAR in os.environ:
        del os.environ[COG_TRAIN_TYPE_STUB_ENV_VAR]
    config = Config(config={"train": train_ref})
    config_train_ref = config.predictor_train_ref
    assert config_train_ref == train_ref, (
        "Train Reference should come from the initial config dictionary."
    )


def test_requires_gpu_env_var():
    gpu = True
    os.environ[COG_GPU_ENV_VAR] = str(gpu)
    config = Config()
    config_gpu = config.requires_gpu
    del os.environ[COG_GPU_ENV_VAR]
    assert config_gpu, "Requires GPU should come from the environment variable."


def test_requires_gpu_no_env_var():
    if COG_GPU_ENV_VAR in os.environ:
        del os.environ[COG_GPU_ENV_VAR]
    config = Config(config={"build": {"gpu": False}})
    config_gpu = config.requires_gpu
    assert not config_gpu, (
        "Requires GPU should come from the initial config dictionary."
    )


def test_get_predictor_ref_predict():
    train_ref = "predict.py:Predictor"
    config = Config(config={"train": train_ref})
    config_train_ref = config.get_predictor_ref(Mode.TRAIN)
    assert train_ref == config_train_ref, (
        "The train ref should equal the config train ref."
    )


def test_get_predictor_ref_train():
    predict_ref = "predict.py:Predictor"
    config = Config(config={"predict": predict_ref})
    config_predict_ref = config.get_predictor_ref(Mode.PREDICT)
    assert predict_ref == config_predict_ref, (
        "The predict ref should equal the config predict ref."
    )


def test_get_predictor_types_with_env_var():
    predict_ref = "predict.py:Predictor"
    os.environ[COG_PREDICT_TYPE_STUB_ENV_VAR] = predict_ref
    os.environ[COG_PREDICT_CODE_STRIP_ENV_VAR] = """
from cog import BasePredictor, Path
from typing import Optional
from pydantic import BaseModel

class ModelOutput(BaseModel):
    success: bool
    error: Optional[str]
    segmentedImage: Optional[Path]

class Predictor(BasePredictor):
    def predict(self, msg: str) -> ModelOutput:
        return None
"""
    config = Config()
    input_type, output_type, is_async = config.get_predictor_types(Mode.PREDICT)
    del os.environ[COG_PREDICT_CODE_STRIP_ENV_VAR]
    del os.environ[COG_PREDICT_TYPE_STUB_ENV_VAR]
    assert str(input_type) == "<class 'cog.predictor.Input'>", (
        "Predict input type should be the predictor Input."
    )
    assert (
        str(output_type) == "<class 'cog.predictor.get_output_type.<locals>.Output'>"
    ), "Predict output type should be the predictor Output."
    assert not is_async, "is_async should be False for normal functions"


def test_get_predictor_types():
    with tempfile.TemporaryDirectory() as tmpdir:
        predict_python_file = os.path.join(tmpdir, "predict.py")
        with open(predict_python_file, "w", encoding="utf-8") as handle:
            handle.write("""
import io

from cog import BasePredictor, Path
from typing import Optional
from pydantic import BaseModel


class ModelOutput(BaseModel):
    success: bool
    error: Optional[str]
    segmentedImage: Optional[Path]


class Predictor(BasePredictor):
    # setup code
    def predict(self, msg: str) -> ModelOutput:
       return ModelOutput(success=False, error=msg, segmentedImage=None)
""")
        predict_ref = f"{predict_python_file}:Predictor"
        config = Config(config={"predict": predict_ref})
        input_type, output_type, is_async = config.get_predictor_types(Mode.PREDICT)
        assert str(input_type) == "<class 'cog.predictor.Input'>", (
            "Predict input type should be the predictor Input."
        )
        assert (
            str(output_type)
            == "<class 'cog.predictor.get_output_type.<locals>.Output'>"
        ), "Predict output type should be the predictor Output."
        assert not is_async, "is_async should be False for normal functions"


def test_get_predictor_types_with_async():
    with tempfile.TemporaryDirectory() as tmpdir:
        predict_python_file = os.path.join(tmpdir, "predict.py")
        with open(predict_python_file, "w", encoding="utf-8") as handle:
            handle.write("""
import io

from cog import BasePredictor, Path
from typing import Optional
from pydantic import BaseModel


class ModelOutput(BaseModel):
    success: bool
    error: Optional[str]
    segmentedImage: Optional[Path]


class Predictor(BasePredictor):
    # setup code
    async def predict(self, msg: str) -> ModelOutput:
       return ModelOutput(success=False, error=msg, segmentedImage=None)
""")
        predict_ref = f"{predict_python_file}:Predictor"
        config = Config(config={"predict": predict_ref})
        input_type, output_type, is_async = config.get_predictor_types(Mode.PREDICT)
        assert str(input_type) == "<class 'cog.predictor.Input'>", (
            "Predict input type should be the predictor Input."
        )
        assert (
            str(output_type)
            == "<class 'cog.predictor.get_output_type.<locals>.Output'>"
        ), "Predict output type should be the predictor Output."
        assert is_async, "is_async should be True for async functions"


def test_get_predictor_types_for_train():
    with tempfile.TemporaryDirectory() as tmpdir:
        predict_python_file = os.path.join(tmpdir, "train.py")
        with open(predict_python_file, "w", encoding="utf-8") as handle:
            handle.write("""
from cog import BaseModel, Input, Path

class TrainingOutput(BaseModel):
    weights: Path

def train(
    n: int,
) -> TrainingOutput:
    with open("weights.bin", "w") as fh:
        for _ in range(n):
            fh.write("a")

    return TrainingOutput(
        weights=Path("weights.bin"),
    )
""")
        train_ref = f"{predict_python_file}:train"
        config = Config(config={"train": train_ref})
        input_type, output_type, is_async = config.get_predictor_types(Mode.TRAIN)
        assert str(input_type) == "<class 'cog.predictor.TrainingInput'>", (
            "Predict input type should be the training Input."
        )
        assert str(output_type).endswith("TrainingOutput'>"), (
            "Predict output type should be the training Output."
        )
        assert not is_async, "is_async should be False for normal functions"


def test_get_predictor_types_for_train_with_async():
    with tempfile.TemporaryDirectory() as tmpdir:
        predict_python_file = os.path.join(tmpdir, "train.py")
        with open(predict_python_file, "w", encoding="utf-8") as handle:
            handle.write("""
from cog import BaseModel, Input, Path

class TrainingOutput(BaseModel):
    weights: Path

async def train(
    n: int,
) -> TrainingOutput:
    with open("weights.bin", "w") as fh:
        for _ in range(n):
            fh.write("a")

    return TrainingOutput(
        weights=Path("weights.bin"),
    )
""")
        train_ref = f"{predict_python_file}:train"
        config = Config(config={"train": train_ref})
        input_type, output_type, is_async = config.get_predictor_types(Mode.TRAIN)
        assert str(input_type) == "<class 'cog.predictor.TrainingInput'>", (
            "Predict input type should be the training Input."
        )
        assert str(output_type).endswith("TrainingOutput'>"), (
            "Predict output type should be the training Output."
        )
        assert is_async, "is_async should be True for async functions"



import io

from cog import BasePredictor, Path
from typing import Optional
from pydantic import BaseModel


class ModelOutput(BaseModel):
    success: bool
    error: Optional[str]
    segmentedImage: Optional[Path]


class Predictor(BasePredictor):
    # setup code
    def predict(self, msg: str) -> ModelOutput:
       return ModelOutput(success=False, error=msg, segmentedImage=None)



import time


class Predictor:
    def setup(self):
        print("did setup")

    def predict(self, steps=5, name="Bob"):
        print("START")
        for i in range(steps):
            time.sleep(0.05)
            print(f"STEP {i+1}")
        print("END")
        return f"NAME={name}"



from cog import BasePredictor, Input, Path


class Predictor(BasePredictor):
    def predict(
        self,
        text: str,
        path: Path,
        num1: int,
        num2: int = Input(default=10),
    ) -> str:
        with open(path) as fh:
            return text + " " + str(num1 * num2) + " " + fh.read()



import os
import signal
import time


def main():
    child_pid = os.fork()
    is_child = child_pid == 0

    pid = os.getpid()
    was_pinged = False

    while True:
        if os.path.exists(".inbox") and is_child:
            s = ""

            with open(".inbox", "r") as inbox:
                print(f"---> CHILD ({pid}) reading request")

                s = inbox.read()

            os.unlink(".inbox")

            with open(".outbox", "w") as outbox:
                print(f"---> CHILD ({pid}) sending response")

                outbox.write("hello " + s)

        if time.time() % 10 == 0:
            if is_child:
                print(f"---> CHILD ({pid}) " + ("here " * 20))
            else:
                print(f"===> PARENT ({pid})")

        time.sleep(0.01)


if __name__ == "__main__":
    main()



import os
import subprocess


def test_config(tmpdir_factory):
    tmpdir = tmpdir_factory.mktemp("project")
    with open(tmpdir / "cog.yaml", "w") as f:
        cog_yaml = """
build:
  python_version: "3.8"
        """
        f.write(cog_yaml)

    subdir = tmpdir / "some/sub/dir"
    os.makedirs(subdir)

    result = subprocess.run(
        ["cog", "run", "echo", "hello world"],
        cwd=subdir,
        check=True,
        capture_output=True,
    )
    assert b"hello world" in result.stdout



import multiprocessing.connection
import multiprocessing.synchronize
import os
import time


def ponger(
    conn: multiprocessing.connection.Connection, lock: multiprocessing.synchronize.Lock
):
    for i in range(100):
        print(f"Getting ready for some serious ponginggg ({i+1}%)")
        time.sleep(0.001 + (0.001 * (i + 1)))

    print("ITS PONGIN TIME")

    pid = os.getpid()

    while True:
        try:
            ping = conn.recv()
            print(f"received {ping} in {pid}")

            with lock:
                print(f"ponging from {pid}")

                conn.send("pong")
                conn.close()

        except EOFError:
            pass



import ast
import json
import sys
import types
import typing
from pathlib import Path

try:
    assert ast.unparse
except (AssertionError, AttributeError):
    # bad "compat" with python3.8
    ast.unparse = repr

BASE_SCHEMA = """
{
  "components": {
    "schemas": {
      "HTTPValidationError": {
        "properties": {
          "detail": {
            "items": { "$ref": "#/components/schemas/ValidationError" },
            "title": "Detail",
            "type": "array"
          }
        },
        "title": "HTTPValidationError",
        "type": "object"
      },
      "PredictionRequest": {
        "properties": {
          "created_at": {
            "format": "date-time",
            "title": "Created At",
            "type": "string"
          },
          "id": { "title": "Id", "type": "string" },
          "input": { "$ref": "#/components/schemas/Input" },
          "output_file_prefix": {
            "title": "Output File Prefix",
            "type": "string"
          },
          "webhook": {
            "format": "uri",
            "maxLength": 65536,
            "minLength": 1,
            "title": "Webhook",
            "type": "string"
          },
          "webhook_events_filter": {
            "default": ["start", "output", "logs", "completed"],
            "items": { "$ref": "#/components/schemas/WebhookEvent" },
            "type": "array"
          }
        },
        "title": "PredictionRequest",
        "type": "object"
      },
      "PredictionResponse": {
        "properties": {
          "completed_at": {
            "format": "date-time",
            "title": "Completed At",
            "type": "string"
          },
          "created_at": {
            "format": "date-time",
            "title": "Created At",
            "type": "string"
          },
          "error": { "title": "Error", "type": "string" },
          "id": { "title": "Id", "type": "string" },
          "input": { "$ref": "#/components/schemas/Input" },
          "logs": { "default": "", "title": "Logs", "type": "string" },
          "metrics": { "title": "Metrics", "type": "object" },
          "output": { "$ref": "#/components/schemas/Output" },
          "started_at": {
            "format": "date-time",
            "title": "Started At",
            "type": "string"
          },
          "status": { "$ref": "#/components/schemas/Status" },
          "version": { "title": "Version", "type": "string" }
        },
        "title": "PredictionResponse",
        "type": "object"
      },
      "Status": {
        "description": "An enumeration.",
        "enum": ["starting", "processing", "succeeded", "canceled", "failed"],
        "title": "Status",
        "type": "string"
      },
      "ValidationError": {
        "properties": {
          "loc": {
            "items": { "anyOf": [{ "type": "string" }, { "type": "integer" }] },
            "title": "Location",
            "type": "array"
          },
          "msg": { "title": "Message", "type": "string" },
          "type": { "title": "Error Type", "type": "string" }
        },
        "required": ["loc", "msg", "type"],
        "title": "ValidationError",
        "type": "object"
      },
      "WebhookEvent": {
        "description": "An enumeration.",
        "enum": ["start", "output", "logs", "completed"],
        "title": "WebhookEvent",
        "type": "string"
      }
    }
  },
  "info": { "title": "Cog", "version": "0.1.0" },
  "openapi": "3.0.2",
  "paths": {
    "/": {
      "get": {
        "operationId": "root__get",
        "responses": {
          "200": {
            "content": {
              "application/json": {
                "schema": { "title": "Response Root  Get" }
              }
            },
            "description": "Successful Response"
          }
        },
        "summary": "Root"
      }
    },
    "/health-check": {
      "get": {
        "operationId": "healthcheck_health_check_get",
        "responses": {
          "200": {
            "content": {
              "application/json": {
                "schema": { "title": "Response Healthcheck Health Check Get" }
              }
            },
            "description": "Successful Response"
          }
        },
        "summary": "Healthcheck"
      }
    },
    "/predictions": {
      "post": {
        "description": "Run a single prediction on the model",
        "operationId": "predict_predictions_post",
        "parameters": [
          {
            "in": "header",
            "name": "prefer",
            "required": false,
            "schema": { "title": "Prefer", "type": "string" }
          }
        ],
        "requestBody": {
          "content": {
            "application/json": {
              "schema": { "$ref": "#/components/schemas/PredictionRequest" }
            }
          }
        },
        "responses": {
          "200": {
            "content": {
              "application/json": {
                "schema": { "$ref": "#/components/schemas/PredictionResponse" }
              }
            },
            "description": "Successful Response"
          },
          "422": {
            "content": {
              "application/json": {
                "schema": { "$ref": "#/components/schemas/HTTPValidationError" }
              }
            },
            "description": "Validation Error"
          }
        },
        "summary": "Predict"
      }
    },
    "/predictions/{prediction_id}": {
      "put": {
        "description": "Run a single prediction on the model (idempotent creation).",
        "operationId": "predict_idempotent_predictions__prediction_id__put",
        "parameters": [
          {
            "in": "path",
            "name": "prediction_id",
            "required": true,
            "schema": { "title": "Prediction ID", "type": "string" }
          },
          {
            "in": "header",
            "name": "prefer",
            "required": false,
            "schema": { "title": "Prefer", "type": "string" }
          }
        ],
        "requestBody": {
          "content": {
            "application/json": {
              "schema": {
                "allOf": [{ "$ref": "#/components/schemas/PredictionRequest" }],
                "title": "Prediction Request"
              }
            }
          },
          "required": true
        },
        "responses": {
          "200": {
            "content": {
              "application/json": {
                "schema": { "$ref": "#/components/schemas/PredictionResponse" }
              }
            },
            "description": "Successful Response"
          },
          "422": {
            "content": {
              "application/json": {
                "schema": { "$ref": "#/components/schemas/HTTPValidationError" }
              }
            },
            "description": "Validation Error"
          }
        },
        "summary": "Predict Idempotent"
      }
    },
    "/predictions/{prediction_id}/cancel": {
      "post": {
        "description": "Cancel a running prediction",
        "operationId": "cancel_predictions__prediction_id__cancel_post",
        "parameters": [
          {
            "in": "path",
            "name": "prediction_id",
            "required": true,
            "schema": { "title": "Prediction ID", "type": "string" }
          }
        ],
        "responses": {
          "200": {
            "content": {
              "application/json": {
                "schema": {
                  "title": "Response Cancel Predictions  Prediction Id  Cancel Post"
                }
              }
            },
            "description": "Successful Response"
          },
          "422": {
            "content": {
              "application/json": {
                "schema": { "$ref": "#/components/schemas/HTTPValidationError" }
              }
            },
            "description": "Validation Error"
          }
        },
        "summary": "Cancel"
      }
    },
    "/shutdown": {
      "post": {
        "operationId": "start_shutdown_shutdown_post",
        "responses": {
          "200": {
            "content": {
              "application/json": {
                "schema": { "title": "Response Start Shutdown Shutdown Post" }
              }
            },
            "description": "Successful Response"
          }
        },
        "summary": "Start Shutdown"
      }
    }
  }
}
"""

OPENAPI_TYPES = {
    "str": "string",  # includes dates, files
    "int": "integer",
    "float": "number",
    "bool": "boolean",
    "list": "array",
    "cog.Path": "string",
    "cog.File": "string",
    "Path": "string",
    "File": "string",
}


def find(obj: ast.AST, name: str) -> ast.AST:
    """Find a particular named node in a tree"""
    return next(node for node in ast.walk(obj) if getattr(node, "name", "") == name)


if typing.TYPE_CHECKING:
    AstVal: "typing.TypeAlias" = (
        "int | float | complex | str | list[AstVal] | bytes | None"
    )
    AstValNoBytes: "typing.TypeAlias" = "int | float | str | list[AstValNoBytes]"
    JSONObject: "typing.TypeAlias" = (
        "int | float | str | list[JSONObject] | JSONDict | None"
    )
    JSONDict: "typing.TypeAlias" = "dict[str, JSONObject]"


def to_serializable(val: "AstVal") -> "JSONObject":
    if isinstance(val, bytes):
        return val.decode("utf-8")
    if isinstance(val, list):
        return [to_serializable(x) for x in val]
    if isinstance(val, complex):
        msg = "complex inputs are not supported"
        raise ValueError(msg)
    return val


def get_value(node: ast.AST) -> "AstVal":
    """Return the value of constant or list of constants"""
    if isinstance(node, ast.Constant):
        return node.value
    if isinstance(node, (ast.List, ast.Tuple)):
        return [get_value(e) for e in node.elts]
    if isinstance(node, ast.UnaryOp) and isinstance(node.op, ast.USub):
        return -typing.cast(typing.Union[int, float, complex], get_value(node.operand))
    if isinstance(node, ast.Index):
        return node.value  # type: ignore
    raise ValueError("Unexpected node type", type(node))


def get_annotation(node: "ast.AST | None") -> str:
    """Return the annotation as a string"""
    if isinstance(node, ast.Name):
        return node.id
    if isinstance(node, ast.Constant):
        return node.value  # e.g. arg: "Path"
    if isinstance(node, ast.Subscript):
        value = get_annotation(node.value)
        if value == "Literal":
            if sys.version_info < (3, 9):
                if isinstance(node.slice, ast.Index):
                    elts = [node.slice.value]
                else:
                    elts = node.slice.elts
            else:
                elts = (
                    node.slice.elts
                    if isinstance(node.slice, ast.Tuple)
                    else [node.slice]
                )
            return f"Literal[{','.join(repr(get_value(e)) for e in elts)}]"
        # ignore other Subscript (Optional[str]), BinOp (str | int), and stuff like that
    raise ValueError("Unexpected annotation type", type(node))


def get_call_name(call: ast.Call) -> str:
    """Try to get the name of a Call"""
    if isinstance(call.func, ast.Name):
        return call.func.id
    if isinstance(call.func, ast.Attribute):
        return call.func.attr
    raise ValueError("Unexpected node type", type(call), ast.unparse(call))


def parse_args(tree: ast.AST) -> "list[tuple[ast.arg, ast.expr | types.EllipsisType]]":
    """Parse argument, default pairs from a file with a predict function"""
    predict = find(tree, "predict")
    assert isinstance(predict, ast.FunctionDef)
    args = predict.args.args  # [-len(defaults) :]
    # use Ellipsis instead of None here to distinguish a default of None
    defaults = [...] * (len(args) - len(predict.args.defaults)) + predict.args.defaults
    return list(zip(args, defaults))


def parse_assignment(assignment: ast.AST) -> "None | tuple[str, JSONObject]":
    """Parse an assignment into an OpenAPI object property"""
    if isinstance(assignment, ast.AnnAssign):
        assert isinstance(assignment.target, ast.Name)  # shouldn't be an Attribute
        default = {}
        if assignment.value:
            try:
                default = {"default": to_serializable(get_value(assignment.value))}
            except UnicodeDecodeError:
                pass
        return assignment.target.id, {
            "title": assignment.target.id.replace("_", " ").title(),
            "type": OPENAPI_TYPES[get_annotation(assignment.annotation)],
            **default,
        }
    if isinstance(assignment, ast.Assign):
        if len(assignment.targets) == 1 and isinstance(assignment.targets[0], ast.Name):
            value = to_serializable(get_value(assignment.value))
            return assignment.targets[0].id, {
                "title": assignment.targets[0].id.replace("_", " ").title(),
                "type": OPENAPI_TYPES[type(value).__name__],
                "default": value,
            }
        raise ValueError("Unexpected assignment", assignment)
    return None


def parse_class(classdef: ast.AST) -> "JSONDict":
    """Parse a class definition into an OpenAPI object"""
    assert isinstance(classdef, ast.ClassDef)
    properties = {
        assignment[0]: assignment[1]
        for assignment in map(parse_assignment, classdef.body)
        if assignment
    }
    return {
        "title": classdef.name,
        "type": "object",
        "properties": properties,
    }


# The supported types are:
# str: a string
# int: an integer
# float: a floating point number
# bool: a boolean
# cog.File: a file-like object representing a file
# cog.Path: a path to a file on disk

BASE_TYPES = ["str", "int", "float", "bool", "File", "Path"]


def resolve_name(node: ast.expr) -> str:
    if isinstance(node, ast.Name):
        return node.id
    if isinstance(node, ast.Index):
        # deprecated, but needed for py3.8
        return resolve_name(node.value)  # type: ignore
    if isinstance(node, ast.Attribute):
        return node.attr
    if isinstance(node, ast.Subscript):
        return resolve_name(node.value)
    raise ValueError("Unexpected node type", type(node), ast.unparse(node))


def parse_return_annotation(
    tree: ast.AST, fn: str = "predict"
) -> "tuple[JSONDict, JSONDict]":
    predict = find(tree, fn)
    if not isinstance(predict, ast.FunctionDef):
        raise ValueError("Could not find predict function")
    annotation = predict.returns
    if not annotation:
        raise TypeError(
            """You must set an output type. If your model can return multiple output types, you can explicitly set `Any` as the output type.

For example:

    from typing import Any

    def predict(
        self,
        image: Path = Input(description="Input image"),
    ) -> Any:
        ...
"""
        )
    # attributes should be resolved to names, maybe blindly
    # subscript values are iterator or
    name = resolve_name(annotation)
    if isinstance(annotation, ast.Subscript):
        # forget about other subscripts like Optional, and assume otherlib.File will still be an uri
        slice = resolve_name(annotation.slice)  # pylint: disable=redefined-builtin
        format = {"format": "uri"} if slice in ("Path", "File") else {}  # pylint: disable=redefined-builtin
        array_type = {"x-cog-array-type": "iterator"} if "Iterator" in name else {}
        display_type = (
            {"x-cog-array-display": "concatenate"} if "Concatenate" in name else {}
        )
        return {}, {
            "title": "Output",
            "type": "array",
            "items": {
                "type": OPENAPI_TYPES.get(slice, slice),
                **format,
            },
            **array_type,
            **display_type,
        }
    if name in BASE_TYPES:
        # otherwise figure this out...
        format = {"format": "uri"} if name in ("Path", "File") else {}
        return {}, {"title": "Output", "type": OPENAPI_TYPES.get(name, name), **format}
    # it must be a custom object
    schema: JSONDict = {name: parse_class(find(tree, name))}
    return schema, {
        "title": "Output",
        "$ref": f"#/components/schemas/{name}",
    }


KEPT_ATTRS = ("description", "default", "ge", "le", "max_length", "min_length", "regex")


def extract_info(code: str) -> "JSONDict":  # pylint: disable=too-many-branches,too-many-locals
    """Parse the schemas from a file with a predict function"""
    tree = ast.parse(code)
    properties: JSONDict = {}
    inputs: JSONDict = {"title": "Input", "type": "object", "properties": properties}
    required: list[str] = []
    schemas: JSONDict = {}
    for arg, default in parse_args(tree):
        if arg.arg == "self":
            continue
        if isinstance(default, ast.Call) and get_call_name(default) == "Input":
            kws = {}
            for kw in default.keywords:
                if kw.arg is None:
                    msg = "unknown argument for Input"
                    raise ValueError(msg)
                kws[kw.arg] = to_serializable(get_value(kw.value))
        elif isinstance(default, (ast.Constant, ast.List, ast.Tuple)):
            kws = {"default": to_serializable(get_value(default))}  # could be None
        elif default == ...:  # no default
            kws = {}
        else:
            raise ValueError("Unexpected default value", default)
        input: JSONDict = {"x-order": len(properties)}
        # need to handle other types?

        annotation = get_annotation(arg.annotation)
        arg_type = OPENAPI_TYPES.get(annotation, "string")
        if annotation in ("Path", "File"):
            input["format"] = "uri"
        elif annotation.startswith("Literal["):
            input["enum"] = list(
                ast.literal_eval(annotation[7:])  # Safely eval the literal values
            )
            arg_type = (
                OPENAPI_TYPES.get(type(input["enum"][0]).__name__, "string")
                if input["enum"]
                else "string"
            )
        for attr in KEPT_ATTRS:
            if attr in kws:
                input[attr] = kws[attr]
        if "default" not in input:
            required.append(arg.arg)
        if "choices" in kws and isinstance(kws["choices"], list):
            input["allOf"] = [{"$ref": f"#/components/schemas/{arg.arg}"}]
            # could use type(kws["choices"][0]).__name__
            schemas[arg.arg] = {
                "title": arg.arg,
                "enum": kws["choices"],
                "type": arg_type,
                "description": "An enumeration.",
            }
        elif "enum" in input:
            input["title"] = arg.arg.replace("_", " ").title()
            input["type"] = arg_type
            input["description"] = "An enumeration."
        else:
            input["title"] = arg.arg.replace("_", " ").title()
            input["type"] = arg_type
        properties[arg.arg] = input
    if required:
        inputs["required"] = list(required)
    # List[Path], list[Path], str, Iterator[str], MyOutput, Output
    return_schema, output = parse_return_annotation(tree, "predict")
    schema: JSONDict = json.loads(BASE_SCHEMA)
    components: JSONDict = {
        "Input": inputs,
        "Output": output,
        **schemas,
        **return_schema,
    }
    # trust me, typechecker, I know BASE_SCHEMA
    x: JSONDict = schema["components"]["schemas"]  # type: ignore
    x.update(components)
    return schema


def extract_file(fname: "str | Path") -> "JSONObject":
    return extract_info(open(fname, encoding="utf-8").read())


if __name__ == "__main__":
    if len(sys.argv) > 1:
        p = Path(sys.argv[1])
        if p.exists():
            print(json.dumps(extract_file(p)))
    else:
        print(json.dumps(extract_info(sys.stdin.read())))



import logging
import os
from pathlib import Path
from typing import Union

log = logging.getLogger(__name__)

PathLike = Union[Path, str, None]


class ProbeHelper:
    _root = Path("/var/run/cog")
    _enabled = False

    def __init__(self, root: PathLike = None) -> None:
        if "KUBERNETES_SERVICE_HOST" not in os.environ:
            log.info("Not running in Kubernetes: disabling probe helpers.")
            return

        if root is not None:
            self._root = Path(root)

        try:
            self._root.mkdir(exist_ok=True, parents=True)
        except OSError:
            log.error(
                "Failed to create cog runtime state directory (%s). "
                "Does it already exist and is a file? Does the user running cog "
                "have permissions?",
                self._root,
            )
        else:
            self._enabled = True

    def ready(self) -> None:
        if self._enabled:
            (self._root / "ready").touch()



from cog import BasePredictor, Path

from typing import List, Union


class Predictor(BasePredictor):
    def predict(self, args: Union[str, List[str]]) -> str:
        if isinstance(args, str):
            return args
        else:
            return "".join(args)



import signal
import subprocess
import sys

from cog import BasePredictor


class Predictor(BasePredictor):
    """
    This predictor checks the case where a process is spawned during setup and then each
    prediction causes that process to write to stdout. In the event that stream
    redirection is not working correctly, the forked process will not be able to write to
    stdout/stderr and will likely exit. Any state other than "running" is considered an
    error condition and raises SystemExit to interrupt any more prediction serving.

    This variant runs a simple subprocess to which SIGUSR1 is sent during each call to
    `predict`.
    """

    def setup(self) -> None:
        print("---> starting background process")

        self.bg = subprocess.Popen(["bash", "child.sh"])

        print(f"---> started background process pid={self.bg.pid}")

    def predict(self, s: str) -> str:
        status = self.bg.poll()

        if status is None:
            print(f"---> sending signal to background job pid={self.bg.pid}")

            self.bg.send_signal(signal.SIGUSR1)

            print(f"---> sent signal to background job pid={self.bg.pid}")
        else:
            print(f"---> background job died status={status}")

            raise SystemExit

        return "hello " + s



import asyncio
import io
import threading
import traceback
import uuid
from abc import ABC, abstractmethod
from concurrent.futures import Future
from datetime import datetime, timezone
from typing import Any, Callable, Dict, Generic, List, Literal, Optional, TypeVar, Union

import requests
import structlog
from attrs import define, field
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from .. import schema
from ..base_input import BaseInput
from ..files import put_file_to_signed_endpoint
from ..json import upload_files
from ..types import PYDANTIC_V2
from .errors import FileUploadError, RunnerBusyError, UnknownPredictionError
from .eventtypes import (
    Done,
    Log,
    PredictionMetric,
    PredictionOutput,
    PredictionOutputType,
)

if PYDANTIC_V2:
    from .helpers import unwrap_pydantic_serialization_iterators
from .telemetry import current_trace_context
from .useragent import get_user_agent
from .webhook import SKIP_START_EVENT, webhook_caller_filtered
from .worker import Worker, _PublicEventType

log = structlog.get_logger("cog.server.runner")


@define
class SetupResult:
    started_at: datetime
    completed_at: Optional[datetime] = None
    logs: List[str] = field(factory=list)
    status: Optional[Literal[schema.Status.FAILED, schema.Status.SUCCEEDED]] = None

    def to_dict(self) -> Dict[str, Any]:
        return {
            "started_at": self.started_at,
            "completed_at": self.completed_at,
            "logs": "".join(self.logs),
            "status": self.status,
        }


class PredictionRunner:
    """
    PredictionRunner manages the state of predictions running through the
    passed worker.
    """

    def __init__(
        self,
        *,
        max_concurrency: int = 1,
        worker: Worker,
    ) -> None:
        self._worker = worker
        self._max_concurrency = max_concurrency

        self._setup_task: Optional[SetupTask] = None
        self._predict_tasks: Dict[str, PredictTask] = {}
        self._predict_tasks_lock = threading.Lock()

    def setup(self) -> "SetupTask":
        assert self._setup_task is None, "do not call setup twice"

        self._setup_task = SetupTask()

        sid = self._worker.subscribe(self._setup_task.handle_event)
        self._setup_task.track(self._worker.setup())
        self._setup_task.add_done_callback(lambda _: self._worker.unsubscribe(sid))

        return self._setup_task

    def predict(
        self,
        prediction: schema.PredictionRequest,
        task_kwargs: Optional[Dict[str, Any]] = None,
    ) -> "PredictTask":
        self._raise_if_busy()

        task_kwargs = task_kwargs or {}

        tag = prediction.id
        if tag is None:
            tag = uuid.uuid4().hex

        task = PredictTask(prediction, **task_kwargs)

        with self._predict_tasks_lock:
            self._predict_tasks[tag] = task

        if isinstance(prediction.input, BaseInput):
            if PYDANTIC_V2:
                payload = unwrap_pydantic_serialization_iterators(
                    prediction.input.model_dump()
                )
            else:
                payload = prediction.input.dict()
        else:
            payload = prediction.input.copy()

        sid = self._worker.subscribe(task.handle_event, tag=tag)
        task.track(self._worker.predict(payload, tag=tag))
        task.add_done_callback(self._task_done_callback(tag, sid))

        return task

    def _task_done_callback(self, tag: str, sid: int) -> Callable[[Any], None]:
        def _callback(_) -> None:
            self._worker.unsubscribe(sid)
            with self._predict_tasks_lock:
                del self._predict_tasks[tag]

        return _callback

    def get_predict_task(self, id: str) -> Optional["PredictTask"]:
        with self._predict_tasks_lock:
            return self._predict_tasks.get(id, None)

    def is_busy(self) -> bool:
        try:
            self._raise_if_busy()
        except RunnerBusyError:
            return True
        return False

    def cancel(self, prediction_id: str) -> None:
        if not prediction_id:
            raise ValueError("prediction_id is required")
        with self._predict_tasks_lock:
            if (
                prediction_id not in self._predict_tasks
                or self._predict_tasks[prediction_id].done()
            ):
                raise UnknownPredictionError("unknown prediction id")
        self._worker.cancel(tag=prediction_id)

    def _raise_if_busy(self) -> None:
        if self._setup_task is None:
            # Setup hasn't been called yet.
            raise RunnerBusyError("setup has not started")
        if not self._setup_task.done():
            # Setup is still running.
            raise RunnerBusyError("setup is not complete")

        with self._predict_tasks_lock:
            processing_tasks = [
                id for id in self._predict_tasks if not self._predict_tasks[id].done()
            ]

        if len(processing_tasks) >= self._max_concurrency:
            # We're at max concurrency
            if self._max_concurrency == 1:
                raise RunnerBusyError("prediction running")
            raise RunnerBusyError("max predictions running")


T = TypeVar("T")


class Task(ABC, Generic[T]):
    @abstractmethod
    def track(self, fut: "Future[Done]") -> None:
        raise NotImplementedError

    @abstractmethod
    def add_done_callback(self, fn: Callable[[T], None]) -> None:
        raise NotImplementedError

    @abstractmethod
    def done(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    def wait(self, timeout: Optional[float] = None) -> None:
        raise NotImplementedError

    @property
    @abstractmethod
    def result(self) -> T:
        raise NotImplementedError


class SetupTask(Task[SetupResult]):
    def __init__(self, _clock: Optional[Callable[[], datetime]] = None) -> None:
        log.info("starting setup")
        self._clock = _clock
        if self._clock is None:
            self._clock = lambda: datetime.now(timezone.utc)

        self._fut: "Optional[Future[Done]]" = None
        self._result = SetupResult(started_at=self._clock())

    @property
    def result(self) -> SetupResult:
        return self._result

    def track(self, fut: "Future[Done]") -> None:
        log.info("started setup")
        self._fut = fut
        self._fut.add_done_callback(self._handle_done)

    def add_done_callback(self, fn: Callable[[SetupResult], None]) -> None:
        assert self._fut, "call track before adding callbacks"
        self._fut.add_done_callback(lambda _: fn(self.result))

    def done(self) -> bool:
        assert self._fut, "call track before checking done"
        return self._fut.done()

    def wait(self, timeout: Optional[float] = None) -> None:
        assert self._fut, "call track before waiting"
        self._fut.result(timeout=timeout)

    def append_logs(self, message: str) -> None:
        self._result.logs.append(message)

    def succeeded(self) -> None:
        log.info("setup succeeded")
        assert self._clock
        self._result.completed_at = self._clock()
        self._result.status = schema.Status.SUCCEEDED

    def failed(self) -> None:
        log.info("setup failed")
        assert self._clock
        self._result.completed_at = self._clock()
        self._result.status = schema.Status.FAILED

    def handle_event(self, event: _PublicEventType) -> None:
        if isinstance(event, Log):
            self.append_logs(event.message)
        elif isinstance(event, Done):
            if event.error:
                self.failed()
            else:
                self.succeeded()
        else:
            log.warn("received unexpected event during setup", data=event)

    def _handle_done(self, f: "Future[Done]") -> None:
        try:
            # See if the future captured an exception...
            f.result()
        except Exception:  # pylint: disable=broad-exception-caught
            log.error("caught exception while running setup", exc_info=True)
            self.append_logs(traceback.format_exc())
            self.failed()


def generate_file_uploader(
    upload_url: str, prediction_id: Optional[str]
) -> Callable[[Any], Any]:
    client = _make_file_upload_http_client()

    def file_uploader(output: Any) -> Any:
        def upload_file(fh: io.IOBase) -> str:
            return put_file_to_signed_endpoint(
                fh, endpoint=upload_url, prediction_id=prediction_id, client=client
            )

        return upload_files(output, upload_file=upload_file)

    return file_uploader


class PredictTask(Task[schema.PredictionResponse]):
    def __init__(
        self,
        prediction_request: schema.PredictionRequest,
        upload_url: Optional[str] = None,
    ) -> None:
        self._log = log.bind(prediction_id=prediction_request.id)

        self._log.info("starting prediction")

        self._fut: "Optional[Future[Done]]" = None

        if PYDANTIC_V2:
            request_dict = unwrap_pydantic_serialization_iterators(
                prediction_request.model_dump()
            )
        else:
            request_dict = prediction_request.dict()

        self._p = schema.PredictionResponse(**request_dict)
        self._p.status = schema.Status.PROCESSING
        self._output_type_multi = None
        self._p.output = None
        self._p.logs = ""
        self._p.started_at = datetime.now(tz=timezone.utc)

        self._webhook_sender = None
        if prediction_request.webhook:
            self._webhook_sender = webhook_caller_filtered(
                str(prediction_request.webhook),
                set(
                    prediction_request.webhook_events_filter
                    or schema.WebhookEvent.default_events()
                ),
            )

        self._file_uploader = None
        if upload_url:
            self._file_uploader = generate_file_uploader(
                upload_url, prediction_id=self._p.id
            )

    @property
    def result(self) -> schema.PredictionResponse:
        return self._p

    def track(self, fut: "Future[Done]") -> None:
        self._log.info("started prediction")

        # HACK: don't send an initial webhook if we're trying to optimize for
        # latency (this guarantees that the first output webhook won't be
        # throttled.)
        if not SKIP_START_EVENT:
            self._send_webhook(schema.WebhookEvent.START)

        self._fut = fut
        self._fut.add_done_callback(self._handle_done)

    def add_done_callback(
        self, fn: Callable[[schema.PredictionResponse], None]
    ) -> None:
        assert self._fut, "call track before adding callbacks"
        self._fut.add_done_callback(lambda _: fn(self.result))

    def done(self) -> bool:
        assert self._fut, "call track before checking done"
        return self._fut.done()

    async def wait_async(self) -> None:
        assert self._fut, "call track before waiting"
        await asyncio.wrap_future(self._fut)
        return None

    def wait(self, timeout: Optional[float] = None) -> None:
        assert self._fut, "call track before waiting"
        self._fut.result(timeout=timeout)

    def set_output_type(self, *, multi: bool) -> None:
        assert self._output_type_multi is None, (
            "Predictor unexpectedly returned multiple output types"
        )
        assert self._p.output is None, (
            "Predictor unexpectedly returned output type after output"
        )

        if multi:
            self._p.output = []

        self._output_type_multi = multi

    def append_output(self, output: Any) -> None:
        assert self._output_type_multi is not None, (
            "Predictor unexpectedly returned output before output type"
        )

        uploaded_output = self._upload_files(output)
        if self._output_type_multi:
            self._p.output.append(uploaded_output)
            self._send_webhook(schema.WebhookEvent.OUTPUT)
        else:
            self._p.output = uploaded_output
            # We don't send a webhook for compatibility with the behaviour of
            # redis_queue. In future we can consider whether it makes sense to send
            # one here.

    def append_logs(self, logs: str) -> None:
        assert self._p.logs is not None
        self._p.logs += logs
        self._send_webhook(schema.WebhookEvent.LOGS)

    def set_metric(self, key: str, value: Union[float, int]) -> None:
        if self._p.metrics is None:
            self._p.metrics = {}
        self._p.metrics[key] = value

    def succeeded(self) -> None:
        self._log.info("prediction succeeded")
        self._p.status = schema.Status.SUCCEEDED
        self._set_completed_at()
        # These have been set already: this is to convince the typechecker of
        # that...
        assert self._p.completed_at is not None
        assert self._p.started_at is not None
        self.set_metric(
            "predict_time",
            (self._p.completed_at - self._p.started_at).total_seconds(),
        )
        self._send_webhook(schema.WebhookEvent.COMPLETED)

    def failed(self, error: str) -> None:
        self._log.info("prediction failed", error=error)
        self._p.status = schema.Status.FAILED
        self._p.error = error
        self._set_completed_at()
        self._send_webhook(schema.WebhookEvent.COMPLETED)

    def canceled(self) -> None:
        self._log.info("prediction canceled")
        self._p.status = schema.Status.CANCELED
        self._set_completed_at()
        self._send_webhook(schema.WebhookEvent.COMPLETED)

    def handle_event(self, event: _PublicEventType) -> None:
        try:
            if isinstance(event, Log):
                self.append_logs(event.message)
            elif isinstance(event, PredictionMetric):
                self.set_metric(event.name, event.value)
            elif isinstance(event, PredictionOutputType):
                self.set_output_type(multi=event.multi)
            elif isinstance(event, PredictionOutput):
                self.append_output(event.payload)
            elif isinstance(event, Done):  # pyright: ignore reportUnnecessaryIsinstance
                if event.canceled:
                    self.canceled()
                elif event.error:
                    self.failed(error=str(event.error_detail))
                else:
                    self.succeeded()
            else:  # shouldn't happen, exhausted the type
                self._log.warn("received unexpected event during predict", data=event)
        except Exception as e:
            self.failed(str(e))

    def _set_completed_at(self) -> None:
        self._p.completed_at = datetime.now(tz=timezone.utc)

    def _send_webhook(self, event: schema.WebhookEvent) -> None:
        if self._webhook_sender is not None:
            self._webhook_sender(self._p, event)

    def _upload_files(self, output: Any) -> Any:
        if self._file_uploader is None:
            return output

        try:
            # TODO: clean up output files
            return self._file_uploader(output)
        except (FileNotFoundError, NotADirectoryError):
            # These error cases indicate that an output path returned by a prediction does
            # not actually exist, so there is no way for us to even attempt to upload it.
            # The error is re-raised without wrapping because this is not considered an
            # "infrastructure error", such as happens during an upload of a file that
            # **does** exist.
            raise
        except Exception as error:  # pylint: disable=broad-exception-caught
            # Any other errors that occur during file upload are irrecoverable and
            # considered "infrastructure errors" because there is a high likelihood that
            # the error happened in a layer that is outside the control of the model.
            raise FileUploadError("Got error trying to upload output files") from error

    def _handle_done(self, f: "Future[Done]") -> None:
        try:
            # See if the future captured an exception...
            f.result()
        except Exception as e:  # pylint: disable=broad-exception-caught
            self._log.error("caught exception while running predict", exc_info=True)
            self.append_logs(traceback.format_exc())
            self.failed(error=str(e))
            self._p._fatal_exception = e


def _make_file_upload_http_client() -> requests.Session:
    session = requests.Session()
    session.headers["user-agent"] = (
        get_user_agent() + " " + str(session.headers["user-agent"])
    )

    ctx = current_trace_context() or {}
    for key, value in ctx.items():
        session.headers[key] = str(value)

    adapter = HTTPAdapter(
        max_retries=Retry(
            total=3,
            backoff_factor=0.1,
            status_forcelist=[408, 429, 500, 502, 503, 504],
            allowed_methods=["PUT"],
        ),
    )
    session.mount("http://", adapter)
    session.mount("https://", adapter)
    return session



from cog import BasePredictor, Input


class Predictor(BasePredictor):
    def setup(self):
        self.prefix = "hello"

    def predict(
        self,
        text: str | None = Input(
            description="Text to prefix with 'hello '", default=None
        ),
    ) -> str:
        return self.prefix + " " + text



import argparse
import asyncio
import functools
import logging
import os
import signal
import socket
import sys
import textwrap
import threading
import traceback
from datetime import datetime, timezone
from enum import Enum, auto, unique
from typing import TYPE_CHECKING, Any, Awaitable, Callable, Dict, Optional, Type

import structlog
import uvicorn
from fastapi import Body, FastAPI, Header, Path, Response
from fastapi.encoders import jsonable_encoder
from fastapi.exceptions import HTTPException
from fastapi.openapi.utils import get_openapi
from fastapi.responses import JSONResponse
from pydantic import ValidationError

from .. import schema
from ..config import Config
from ..errors import PredictorNotSet
from ..files import upload_file
from ..json import upload_files
from ..logging import setup_logging
from ..mode import Mode
from ..types import PYDANTIC_V2

try:
    from .._version import __version__
except ImportError:
    __version__ = "dev"

if PYDANTIC_V2:
    from .helpers import (
        unwrap_pydantic_serialization_iterators,
        update_openapi_schema_for_pydantic_2,
    )

from .probes import ProbeHelper
from .runner import (
    PredictionRunner,
    RunnerBusyError,
    SetupResult,
    UnknownPredictionError,
)
from .telemetry import make_trace_context, trace_context
from .worker import make_worker

if TYPE_CHECKING:
    from typing import ParamSpec, TypeVar  # pylint: disable=import-outside-toplevel

    P = ParamSpec("P")  # pylint: disable=invalid-name
    T = TypeVar("T")  # pylint: disable=invalid-name

log = structlog.get_logger("cog.server.http")


@unique
class Health(Enum):
    UNKNOWN = auto()
    STARTING = auto()
    READY = auto()
    BUSY = auto()
    SETUP_FAILED = auto()
    DEFUNCT = auto()


class MyState:
    health: Health
    setup_result: Optional[SetupResult]


class MyFastAPI(FastAPI):
    # TODO: not, strictly speaking, legal
    # https://github.com/microsoft/pyright/issues/5933
    # but it'd need a FastAPI patch to fix
    state: MyState  # type: ignore


def add_setup_failed_routes(
    app: MyFastAPI,  # pylint: disable=redefined-outer-name
    started_at: datetime,
    msg: str,
) -> None:
    print(msg)
    result = SetupResult(
        started_at=started_at,
        completed_at=datetime.now(tz=timezone.utc),
        logs=[msg],
        status=schema.Status.FAILED,
    )
    app.state.setup_result = result
    app.state.health = Health.SETUP_FAILED

    @app.get("/health-check")
    async def healthcheck_startup_failed() -> Any:
        assert app.state.setup_result
        return jsonable_encoder(
            {
                "status": app.state.health.name,
                "setup": app.state.setup_result.to_dict(),
            }
        )


def create_app(  # pylint: disable=too-many-arguments,too-many-locals,too-many-statements
    cog_config: Config,
    shutdown_event: Optional[threading.Event],  # pylint: disable=redefined-outer-name
    app_threads: Optional[int] = None,
    upload_url: Optional[str] = None,
    mode: Mode = Mode.PREDICT,
    is_build: bool = False,
    await_explicit_shutdown: bool = False,  # pylint: disable=redefined-outer-name
) -> MyFastAPI:
    app = MyFastAPI(  # pylint: disable=redefined-outer-name
        title="Cog",  # TODO: mention model name?
        # version=None # TODO
    )

    def custom_openapi() -> Dict[str, Any]:
        if not app.openapi_schema:
            openapi_schema = get_openapi(
                title="Cog",
                openapi_version="3.0.2",
                version="0.1.0",
                routes=app.routes,
            )

            # Pydantic 2 changes how optional fields are represented in OpenAPI schema.
            # See: https://github.com/tiangolo/fastapi/pull/9873#issuecomment-1997105091
            if PYDANTIC_V2:
                update_openapi_schema_for_pydantic_2(openapi_schema)

            app.openapi_schema = openapi_schema

        return app.openapi_schema

    app.openapi = custom_openapi

    app.state.health = Health.STARTING
    app.state.setup_result = None
    started_at = datetime.now(tz=timezone.utc)

    # shutdown is needed no matter what happens
    @app.post("/shutdown")
    async def start_shutdown() -> Any:
        log.info("shutdown requested via http")
        if shutdown_event:
            shutdown_event.set()
        return JSONResponse({}, status_code=200)

    try:
        InputType, OutputType, is_async = cog_config.get_predictor_types(
            mode=Mode.PREDICT
        )
    except Exception:  # pylint: disable=broad-exception-caught
        msg = "Error while loading predictor:\n\n" + traceback.format_exc()
        add_setup_failed_routes(app, started_at, msg)
        return app

    worker = make_worker(
        predictor_ref=cog_config.get_predictor_ref(mode=mode),
        is_async=is_async,
        max_concurrency=cog_config.max_concurrency,
    )
    runner = PredictionRunner(worker=worker, max_concurrency=cog_config.max_concurrency)

    class PredictionRequest(schema.PredictionRequest.with_types(input_type=InputType)):
        pass

    PredictionResponse = schema.PredictionResponse.with_types(  # pylint: disable=invalid-name
        input_type=InputType, output_type=OutputType
    )

    if app_threads is None:
        app_threads = 1 if cog_config.requires_gpu else _cpu_count()
    http_semaphore = asyncio.Semaphore(app_threads)

    def limited(f: "Callable[P, Awaitable[T]]") -> "Callable[P, Awaitable[T]]":
        @functools.wraps(f)
        async def wrapped(*args: "P.args", **kwargs: "P.kwargs") -> "T":  # pylint: disable=redefined-outer-name
            async with http_semaphore:
                return await f(*args, **kwargs)

        return wrapped

    index_document = {
        "cog_version": __version__,
        "docs_url": "/docs",
        "openapi_url": "/openapi.json",
        "shutdown_url": "/shutdown",
        "healthcheck_url": "/health-check",
        "predictions_url": "/predictions",
        "predictions_idempotent_url": "/predictions/{prediction_id}",
        "predictions_cancel_url": "/predictions/{prediction_id}/cancel",
    }

    if cog_config.predictor_train_ref:
        try:
            TrainingInputType, TrainingOutputType, _ = cog_config.get_predictor_types(
                Mode.TRAIN
            )

            class TrainingRequest(
                schema.TrainingRequest.with_types(input_type=TrainingInputType)
            ):
                pass

            TrainingResponse = schema.TrainingResponse.with_types(  # pylint: disable=invalid-name
                input_type=TrainingInputType, output_type=TrainingOutputType
            )

            @app.post(
                "/trainings",
                response_model=TrainingResponse,
                response_model_exclude_unset=True,
            )
            async def train(
                request: TrainingRequest = Body(default=None),
                prefer: Optional[str] = Header(default=None),
                traceparent: Optional[str] = Header(
                    default=None, include_in_schema=False
                ),
                tracestate: Optional[str] = Header(
                    default=None, include_in_schema=False
                ),
            ) -> Any:  # type: ignore
                respond_async = prefer == "respond-async"

                with trace_context(make_trace_context(traceparent, tracestate)):
                    return await _predict(
                        request=request,
                        response_type=TrainingResponse,
                        respond_async=respond_async,
                    )

            @app.put(
                "/trainings/{training_id}",
                response_model=TrainingResponse,
                response_model_exclude_unset=True,
            )
            async def train_idempotent(
                training_id: str = Path(..., title="Training ID"),
                request: TrainingRequest = Body(..., title="Training Request"),
                prefer: Optional[str] = Header(default=None),
                traceparent: Optional[str] = Header(
                    default=None, include_in_schema=False
                ),
                tracestate: Optional[str] = Header(
                    default=None, include_in_schema=False
                ),
            ) -> Any:
                if request.id is not None and request.id != training_id:
                    body = {
                        "loc": ("body", "id"),
                        "msg": "training ID must match the ID supplied in the URL",
                        "type": "value_error",
                    }
                    raise HTTPException(422, [body])

                # We've already checked that the IDs match, now ensure that an ID is
                # set on the prediction object
                request.id = training_id

                # If the prediction service is already running a prediction with a
                # matching ID, return its current state.
                if runner.is_busy():
                    task = runner.get_predict_task(request.id)
                    if task:
                        return JSONResponse(
                            jsonable_encoder(task.result),
                            status_code=202,
                        )

                # TODO: spec-compliant parsing of Prefer header.
                respond_async = prefer == "respond-async"

                with trace_context(make_trace_context(traceparent, tracestate)):
                    return await _predict(
                        request=request,
                        response_type=TrainingResponse,
                        respond_async=respond_async,
                    )

            @app.post("/trainings/{training_id}/cancel")
            def cancel_training(
                training_id: str = Path(..., title="Training ID"),
            ) -> Any:
                return cancel(training_id)

            index_document.update(
                {
                    "trainings_url": "/trainings",
                    "trainings_idempotent_url": "/trainings/{training_id}",
                    "trainings_cancel_url": "/trainings/{training_id}/cancel",
                }
            )

        except Exception as e:  # pylint: disable=broad-exception-caught
            if isinstance(e, (PredictorNotSet, FileNotFoundError)) and not is_build:
                pass  # ignore missing train.py for backward compatibility with existing "bad" models in use
            else:
                app.state.health = Health.SETUP_FAILED
                msg = "Error while loading trainer:\n\n" + traceback.format_exc()
                add_setup_failed_routes(app, started_at, msg)
                return app

    @app.on_event("startup")
    def startup() -> None:
        # check for early setup failures
        if (
            app.state.setup_result
            and app.state.setup_result.status == schema.Status.FAILED
        ):
            # signal shutdown if interactive run
            if shutdown_event and not await_explicit_shutdown:
                shutdown_event.set()
        else:
            setup_task = runner.setup()
            setup_task.add_done_callback(_handle_setup_done)

    @app.on_event("shutdown")
    def shutdown() -> None:
        worker.terminate()

    @app.get("/")
    async def root() -> Any:
        return index_document

    @app.get("/health-check")
    async def healthcheck() -> Any:
        if app.state.health == Health.READY:
            health = Health.BUSY if runner.is_busy() else Health.READY
        else:
            health = app.state.health
        setup = app.state.setup_result.to_dict() if app.state.setup_result else {}
        return jsonable_encoder({"status": health.name, "setup": setup})

    @limited
    @app.post(
        "/predictions",
        response_model=PredictionResponse,
        response_model_exclude_unset=True,
    )
    async def predict(
        request: PredictionRequest = Body(default=None),
        prefer: Optional[str] = Header(default=None),
        traceparent: Optional[str] = Header(default=None, include_in_schema=False),
        tracestate: Optional[str] = Header(default=None, include_in_schema=False),
    ) -> Any:  # type: ignore
        """
        Run a single prediction on the model
        """
        # TODO: spec-compliant parsing of Prefer header.
        respond_async = prefer == "respond-async"

        with trace_context(make_trace_context(traceparent, tracestate)):
            return await _predict(
                request=request,
                response_type=PredictionResponse,
                respond_async=respond_async,
            )

    @limited
    @app.put(
        "/predictions/{prediction_id}",
        response_model=PredictionResponse,
        response_model_exclude_unset=True,
    )
    async def predict_idempotent(
        prediction_id: str = Path(..., title="Prediction ID"),
        request: PredictionRequest = Body(..., title="Prediction Request"),
        prefer: Optional[str] = Header(default=None),
        traceparent: Optional[str] = Header(default=None, include_in_schema=False),
        tracestate: Optional[str] = Header(default=None, include_in_schema=False),
    ) -> Any:
        """
        Run a single prediction on the model (idempotent creation).
        """
        if request.id is not None and request.id != prediction_id:
            body = {
                "loc": ("body", "id"),
                "msg": "prediction ID must match the ID supplied in the URL",
                "type": "value_error",
            }
            raise HTTPException(422, [body])

        # We've already checked that the IDs match, now ensure that an ID is
        # set on the prediction object
        request.id = prediction_id

        # If the prediction service is already running a prediction with a
        # matching ID, return its current state.
        if runner.is_busy():
            task = runner.get_predict_task(request.id)
            if task:
                return JSONResponse(
                    jsonable_encoder(task.result),
                    status_code=202,
                )

        # TODO: spec-compliant parsing of Prefer header.
        respond_async = prefer == "respond-async"

        with trace_context(make_trace_context(traceparent, tracestate)):
            return await _predict(
                request=request,
                response_type=PredictionResponse,
                respond_async=respond_async,
            )

    async def _predict(
        *,
        request: Optional[PredictionRequest],
        response_type: Type[schema.PredictionResponse],
        respond_async: bool = False,
    ) -> Response:
        # [compat] If no body is supplied, assume that this model can be run
        # with empty input. This will throw a ValidationError if that's not
        # possible.
        if request is None:
            request = PredictionRequest(input={})
        # [compat] If body is supplied but input is None, set it to an empty
        # dictionary so that later code can be simpler.
        if request.input is None:
            request.input = {}  # pylint: disable=attribute-defined-outside-init

        task_kwargs = {}
        if respond_async:
            # For now, we only ask PredictionService to handle file uploads for
            # async predictions. This is unfortunate but required to ensure
            # backwards-compatible behaviour for synchronous predictions.
            task_kwargs["upload_url"] = upload_url

        try:
            predict_task = runner.predict(request, task_kwargs=task_kwargs)
        except RunnerBusyError:
            return JSONResponse(
                {"detail": "Already running a prediction"}, status_code=409
            )

        if hasattr(request.input, "cleanup"):
            predict_task.add_done_callback(lambda _: request.input.cleanup())

        predict_task.add_done_callback(_handle_predict_done)

        if respond_async:
            return JSONResponse(
                jsonable_encoder(predict_task.result),
                status_code=202,
            )

        # Otherwise, wait for the prediction to complete...
        await predict_task.wait_async()

        # ...and return the result.
        if PYDANTIC_V2:
            response_object = unwrap_pydantic_serialization_iterators(
                predict_task.result.model_dump()
            )
        else:
            response_object = predict_task.result.dict()
        try:
            _ = response_type(**response_object)
        except ValidationError as e:
            _log_invalid_output(e, mode)
            raise HTTPException(status_code=500, detail=str(e)) from e

        response_object["output"] = upload_files(
            response_object["output"],
            upload_file=lambda fh: upload_file(fh, request.output_file_prefix),  # type: ignore
        )

        # FIXME: clean up output files
        encoded_response = jsonable_encoder(response_object)
        return JSONResponse(content=encoded_response)

    @app.post("/predictions/{prediction_id}/cancel")
    async def cancel(prediction_id: str = Path(..., title="Prediction ID")) -> Any:
        """
        Cancel a running prediction
        """
        try:
            runner.cancel(prediction_id)
        except UnknownPredictionError:
            return JSONResponse({}, status_code=404)
        return JSONResponse({}, status_code=200)

    def _handle_predict_done(response: schema.PredictionResponse) -> None:
        if response._fatal_exception:
            _maybe_shutdown(response._fatal_exception)

    def _handle_setup_done(setup_result: SetupResult) -> None:
        app.state.setup_result = setup_result

        if app.state.setup_result.status == schema.Status.SUCCEEDED:
            app.state.health = Health.READY

            # In kubernetes, mark the pod as ready now setup has completed.
            probes = ProbeHelper()
            probes.ready()
        else:
            _maybe_shutdown(Exception("setup failed"), status=Health.SETUP_FAILED)

    def _maybe_shutdown(exc: BaseException, *, status: Health = Health.DEFUNCT) -> None:
        log.error("encountered fatal error", exc_info=exc)
        app.state.health = status
        if shutdown_event and not await_explicit_shutdown:
            log.error("shutting down immediately")
            shutdown_event.set()
        else:
            log.error("awaiting explicit shutdown")

    return app


def _log_invalid_output(error: Any, mode: Mode) -> None:
    function_name = "predict()"
    if mode == Mode.TRAIN:
        function_name = "train()"
    log.error(
        textwrap.dedent(
            f"""\
            The return value of {function_name} was not valid:

            {error}

            Check that your predict function is in this form, where `output_type` is the same as the type you are returning (e.g. `str`):

                def {function_name} -> output_type:
                    ...
           """
        )
    )


class Server(uvicorn.Server):
    def start(self) -> None:
        self._thread = threading.Thread(target=self.run)  # pylint: disable=attribute-defined-outside-init
        self._thread.start()

    def stop(self) -> None:
        log.info("stopping server")
        self.should_exit = True  # pylint: disable=attribute-defined-outside-init

        self._thread.join(timeout=5)
        if not self._thread.is_alive():
            return

        log.warn("failed to exit after 5 seconds, setting force_exit")
        self.force_exit = True  # pylint: disable=attribute-defined-outside-init
        self._thread.join(timeout=5)
        if not self._thread.is_alive():
            return

        log.warn("failed to exit after another 5 seconds, sending SIGKILL")
        os.kill(os.getpid(), signal.SIGKILL)


def is_port_in_use(port: int) -> bool:  # pylint: disable=redefined-outer-name
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        return sock.connect_ex(("localhost", port)) == 0


def signal_ignore(signum: Any, frame: Any) -> None:  # pylint: disable=unused-argument
    log.warn("Got a signal to exit, ignoring it...", signal=signal.Signals(signum).name)


def signal_set_event(event: threading.Event) -> Callable[[Any, Any], None]:
    def _signal_set_event(signum: Any, frame: Any) -> None:  # pylint: disable=unused-argument
        event.set()

    return _signal_set_event


def _cpu_count() -> int:
    try:
        return len(os.sched_getaffinity(0)) or 1  # type: ignore
    except AttributeError:  # not available on every platform
        return os.cpu_count() or 1


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Cog HTTP server")
    parser.add_argument(
        "-v", "--version", action="store_true", help="Show version and exit"
    )
    parser.add_argument(
        "--host",
        dest="host",
        type=str,
        default="0.0.0.0",
        help="Host to bind to",
    )
    parser.add_argument(
        "--threads",
        dest="threads",
        type=int,
        default=None,
        help="Number of worker processes. Defaults to number of CPUs, or 1 if using a GPU.",
    )
    parser.add_argument(
        "--upload-url",
        dest="upload_url",
        type=str,
        default=None,
        help="An endpoint for Cog to PUT output files to",
    )
    parser.add_argument(
        "--await-explicit-shutdown",
        dest="await_explicit_shutdown",
        type=bool,
        default=False,
        help="Ignore SIGTERM and wait for a request to /shutdown (or a SIGINT) before exiting",
    )
    parser.add_argument(
        "--x-mode",
        dest="mode",
        type=Mode,
        default=Mode.PREDICT,
        choices=list(Mode),
        help="Experimental: Run in 'predict' or 'train' mode",
    )
    args = parser.parse_args()

    if args.version:
        print(f"cog.server.http {__version__}")
        sys.exit(0)

    # log level is configurable so we can make it quiet or verbose for `cog predict`
    # cog predict --debug       # -> debug
    # cog predict               # -> warning
    # docker run <image-name>   # -> info (default)
    log_level = logging.getLevelName(os.environ.get("COG_LOG_LEVEL", "INFO").upper())
    setup_logging(log_level=log_level)

    shutdown_event = threading.Event()

    await_explicit_shutdown = args.await_explicit_shutdown
    if await_explicit_shutdown:
        signal.signal(signal.SIGTERM, signal_ignore)
    else:
        signal.signal(signal.SIGTERM, signal_set_event(shutdown_event))

    app = create_app(
        cog_config=Config(),
        shutdown_event=shutdown_event,
        app_threads=args.threads,
        upload_url=args.upload_url,
        mode=args.mode,
        await_explicit_shutdown=await_explicit_shutdown,
    )

    host: str = args.host

    port = int(os.getenv("PORT", "5000"))
    if is_port_in_use(port):
        log.error(f"Port {port} is already in use")
        sys.exit(1)

    server_config = uvicorn.Config(
        app,
        host=host,
        port=port,
        log_config=None,
        # This is the default, but to be explicit: only run a single worker
        workers=1,
    )

    s = Server(config=server_config)
    s.start()

    try:
        shutdown_event.wait()
    except KeyboardInterrupt:
        pass

    s.stop()

    # return error exit code when setup failed and cog is running in interactive mode (not k8s)
    if (
        app.state.setup_result
        and app.state.setup_result.status == schema.Status.FAILED
        and not await_explicit_shutdown
    ):
        sys.exit(-1)



import inspect
import os
import sys
import uuid

import pytest

from cog.code_xforms import load_module_from_string, strip_model_source_code
from cog.predictor import (
    get_predict,
    get_predictor,
    load_full_predictor_from_file,
)

PREDICTOR_FIXTURES = [
    ("input_choices", "Predictor", "predict"),
    ("input_choices_integer", "Predictor", "predict"),
    ("input_choices_iterable", "Predictor", "predict"),
    ("input_file", "Predictor", "predict"),
    ("function", "predict", "predict"),
    ("input_ge_le", "Predictor", "predict"),
    ("input_integer", "Predictor", "predict"),
    ("input_integer_default", "Predictor", "predict"),
    ("input_multiple", "Predictor", "predict"),
    ("input_none", "Predictor", "predict"),
    ("input_path", "Predictor", "predict"),
    ("input_path_2", "Predictor", "predict"),
    ("input_string", "Predictor", "predict"),
    ("input_union_integer_or_list_of_integers", "Predictor", "predict"),
    ("input_union_string_or_list_of_strings", "Predictor", "predict"),
    ("complex_output", "Predictor", "predict"),
    ("output_complex", "Predictor", "predict"),
    ("output_file_named", "Predictor", "predict"),
    ("output_file", "Predictor", "predict"),
    ("output_path_image", "Predictor", "predict"),
    ("output_path_text", "Predictor", "predict"),
    ("output_numpy", "Predictor", "predict"),
    ("output_iterator_complex", "Predictor", "predict"),
    ("yield_concatenate_iterator", "Predictor", "predict"),
    ("yield_files", "Predictor", "predict"),
    ("yield_strings_file_input", "Predictor", "predict"),
    ("yield_strings", "Predictor", "predict"),
]


def _fixture_path(name):
    test_dir = os.path.dirname(os.path.realpath(__file__))
    return os.path.join(test_dir, f"fixtures/{name}.py")


@pytest.mark.skipif(sys.version_info < (3, 9), reason="Requires Python 3.9 or newer")
@pytest.mark.parametrize("fixture_name, class_name, method_name", PREDICTOR_FIXTURES)
def test_fast_slow_signatures(fixture_name: str, class_name: str, method_name: str):
    module_path = _fixture_path(fixture_name)
    # get signature from FAST loader
    code = None
    with open(module_path, encoding="utf-8") as file:
        code = strip_model_source_code(file.read(), [class_name], [method_name])
    module_fast = load_module_from_string(uuid.uuid4().hex, code)
    assert hasattr(module_fast, class_name)
    predictor_fast = get_predictor(module_fast, class_name)
    predict_fast = get_predict(predictor_fast)
    signature_fast = inspect.signature(predict_fast)
    # get signature from SLOW loader
    module_slow = load_full_predictor_from_file(module_path, module_fast.__name__)
    assert hasattr(module_slow, class_name)
    predictor_slow = get_predictor(module_slow, class_name)
    predict_slow = get_predict(predictor_slow)
    signature_slow = inspect.signature(predict_slow)
    # compare predict signatures using str representation (good enough) as some custom Fields do not have __eq__
    assert str(signature_fast) == str(signature_slow)



import ast
import re
import types
from typing import List, Optional, Set, Tuple, Union

COG_IMPORT_MODULES = {
    "cog",
    "typing",
    "typing_extensions",
    "sys",
    "os",
    "functools",
    "pydantic",
    "numpy",
}


def load_module_from_string(
    name: str, source: Optional[str]
) -> Optional[types.ModuleType]:
    if not source or not name:
        return None
    module = types.ModuleType(name)
    exec(source, module.__dict__)  # noqa: S102 # pylint: disable=exec-used
    return module


def extract_class_sources(source_code: str, class_names: List[str]) -> List[str]:
    """
    Extracts the source code for a specified class from a given source text.
    Args:
        source_code: The complete source code as a string.
        class_name: The name of the class to extract.
    Returns:
        The source code of the specified class if found; otherwise, an empty string.
    """
    class_name_pattern = re.compile(r"\b[A-Z]\w*\b")
    all_class_names = []
    for class_name in class_names:
        all_class_names.extend(class_name_pattern.findall(class_name))

    class ClassExtractor(ast.NodeVisitor):
        def __init__(self) -> None:
            self.class_sources = []

        def visit_ClassDef(self, node: ast.ClassDef) -> None:  # pylint: disable=invalid-name
            self.class_sources.append(node)

    tree = ast.parse(source_code)
    extractor = ClassExtractor()
    extractor.visit(tree)

    valid_class_names = set(all_class_names)
    for node in extractor.class_sources:
        if node.name not in valid_class_names:
            continue
        for base_name in node.bases:
            valid_class_names.add(base_name.id)

    return [
        str(ast.get_source_segment(source_code, x))
        for x in extractor.class_sources
        if x.name in valid_class_names
    ]


def extract_function_source(source_code: str, function_names: List[str]) -> str:
    """
    Extracts the source code for a specified function from a given source text.
    Args:
        source_code: The complete source code as a string.
        function_name: The name of the function to extract.
    Returns:
        The source code of the specified function if found; otherwise, an empty string.
    """

    class FunctionExtractor(ast.NodeVisitor):
        def __init__(self) -> None:
            self.function_sources = []

        def visit_FunctionDef(self, node: ast.FunctionDef) -> None:  # pylint: disable=invalid-name
            if node.name in function_names and not isinstance(node, ast.Module):
                # Extract the source segment for this function definition
                self.function_sources.append(ast.get_source_segment(source_code, node))

    tree = ast.parse(source_code)
    extractor = FunctionExtractor()
    extractor.visit(tree)
    return "\n".join(extractor.function_sources)


def make_class_methods_empty(
    source_code: Union[str, ast.AST],
    class_name: Optional[str],
    global_vars: List[ast.Assign],
) -> Tuple[str, List[ast.Assign]]:
    """
    Transforms the source code of a specified class to remove the bodies of all its methods
    and replace them with 'return None'.
    Args:
        source_code: The complete source code as a string.
        class_name: The name of the class to transform.
    Returns:
        The transformed source code of the specified class.
    """

    class MethodBodyTransformer(ast.NodeTransformer):
        def __init__(self, global_vars: List[ast.Assign]) -> None:
            self.used_globals = set()
            self._targets = {
                target.id: global_name
                for global_name in global_vars
                for target in global_name.targets
                if isinstance(target, ast.Name)
            }

        def visit_ClassDef(self, node: ast.ClassDef) -> Optional[ast.AST]:  # pylint: disable=invalid-name
            if class_name is None or node.name == class_name:
                for body_item in node.body:
                    if isinstance(body_item, ast.FunctionDef):
                        # Replace the body of the method with `return None`
                        body_item.body = [ast.Return(value=ast.Constant(value=None))]
                        # Remove decorators from the function
                        body_item.decorator_list = []
                        # Determine if one our globals is referenced by the function.
                        for default in body_item.args.defaults:
                            if isinstance(default, ast.Call):
                                for keyword in default.keywords:
                                    if isinstance(keyword.value, ast.Name):
                                        corresponding_global = self._targets.get(
                                            keyword.value.id
                                        )
                                        if corresponding_global is not None:
                                            self.used_globals.add(corresponding_global)
                return node

            return None

    tree = source_code if isinstance(source_code, ast.AST) else ast.parse(source_code)
    transformer = MethodBodyTransformer(global_vars)
    transformed_tree = transformer.visit(tree)
    class_code = ast.unparse(transformed_tree)
    return class_code, list(transformer.used_globals)


def extract_method_return_type(
    source_code: Union[str, ast.AST], class_names: List[str], method_names: List[str]
) -> List[str]:
    """
    Extracts the return type annotation of a specified method within a given class from the source code.
    Args:
        source_code: A string containing the Python source code.
        class_name: The name of the class containing the method of interest.
        method_name: The name of the method whose return type annotation is to be extracted.
    Returns:
        A string representing the method's return type annotation if found; otherwise, None.
    """

    class MethodReturnTypeExtractor(ast.NodeVisitor):
        def __init__(self) -> None:
            self.return_types = []

        def visit_ClassDef(self, node: ast.ClassDef) -> None:  # pylint: disable=invalid-name
            if node.name in class_names:
                self.generic_visit(node)

        def visit_FunctionDef(self, node: ast.FunctionDef) -> None:  # pylint: disable=invalid-name
            if node.name in method_names and node.returns:
                self.return_types.append(ast.unparse(node.returns))

    tree = source_code if isinstance(source_code, ast.AST) else ast.parse(source_code)
    extractor = MethodReturnTypeExtractor()
    extractor.visit(tree)

    return extractor.return_types


def extract_function_return_types(
    source_code: Union[str, ast.AST], function_names: List[str]
) -> List[str]:
    """
    Extracts the return type annotation of a specified function from the source code.
    Args:
        source_code: A string containing the Python source code.
        function_name: The name of the function whose return type annotation is to be extracted.
    Returns:
        A string representing the function's return type annotation if found; otherwise, None.
    """

    class FunctionReturnTypeExtractor(ast.NodeVisitor):
        def __init__(self) -> None:
            self.return_types = []

        def visit_FunctionDef(self, node: ast.FunctionDef) -> None:  # pylint: disable=invalid-name
            if node.name in function_names and node.returns:
                # Extract and return the string representation of the return type
                self.return_types.append(ast.unparse(node.returns))

    tree = source_code if isinstance(source_code, ast.AST) else ast.parse(source_code)
    extractor = FunctionReturnTypeExtractor()
    extractor.visit(tree)

    return extractor.return_types


def make_function_empty(
    source_code: Union[str, ast.AST], function_names: List[str]
) -> str:
    """
    Transforms the source code to remove the body of a specified function
    and replace it with 'return None'.
    Args:
        source_code: The complete source code as a string or an AST node.
        function_name: The name of the function to transform.
    Returns:
        The transformed source code with the specified function's body emptied.
    """

    class FunctionBodyTransformer(ast.NodeTransformer):
        def visit_FunctionDef(self, node: ast.FunctionDef) -> Optional[ast.AST]:  # pylint: disable=invalid-name
            if node.name in function_names:
                # Replace the body of the function with `return None`
                node.body = [ast.Return(value=ast.Constant(value=None))]
                return node

            return None

    tree = source_code if isinstance(source_code, ast.AST) else ast.parse(source_code)
    transformer = FunctionBodyTransformer()
    transformed_tree = transformer.visit(tree)
    modified_code = ast.unparse(transformed_tree)
    return modified_code


def extract_specific_imports(
    source_code: Union[str, ast.AST], module_names: Set[str]
) -> str:
    """
    Extracts import statements from the source code that match a specified list of module names.
    Args:
        source_code: The Python source code as a string.
        module_names: A set of module names for which to extract import statements.
    Returns:
        A list of strings, each string is an import statement that matches one of the specified module names.
    """

    class ImportExtractor(ast.NodeVisitor):
        def __init__(self) -> None:
            self.imports = []

        def visit_Import(self, node: ast.Import) -> None:  # pylint: disable=invalid-name
            for alias in node.names:
                if alias.name in module_names:
                    self.imports.append(ast.unparse(node))

        def visit_ImportFrom(self, node: ast.ImportFrom) -> None:  # pylint: disable=invalid-name
            if node.module in module_names:
                self.imports.append(ast.unparse(node))

    tree = source_code if isinstance(source_code, ast.AST) else ast.parse(source_code)
    extractor = ImportExtractor()
    extractor.visit(tree)

    return "\n".join(extractor.imports)


def _extract_globals(source_code: Union[str, ast.AST]) -> List[ast.Assign]:
    tree = source_code if isinstance(source_code, ast.AST) else ast.parse(source_code)
    if isinstance(tree, ast.Module):
        return [x for x in tree.body if isinstance(x, ast.Assign)]
    return []


def _render_globals(global_vars: List[ast.Assign]) -> str:
    return "\n".join([ast.unparse(x) for x in global_vars])


def strip_model_source_code(
    source_code: str, class_names: List[str], method_names: List[str]
) -> Optional[str]:
    """
    Strips down the model source code by extracting relevant classes and making methods empty.
    Args:
        source_code: The complete model source code as a string.
        class_name: The name of the class to be processed. If empty or the class is not found,
                    it falls back to processing a function specified by `method_name`.
        method_name: The name of the method (if processing a class) or the function (if processing standalone functions)
                     whose return type is to be extracted and used in generating the final model source.
    Returns:
        A string containing the modified source code, including a predefined header.
        Returns None if neither the class nor the function specified could be found or processed.
    """
    imports = extract_specific_imports(source_code, COG_IMPORT_MODULES)
    class_sources = (
        None if not class_names else extract_class_sources(source_code, class_names)
    )
    global_vars = _extract_globals(source_code)
    if class_sources:
        class_source = "\n".join(class_sources)
        class_source, global_vars = make_class_methods_empty(
            class_source, None, global_vars
        )
        return_types = extract_method_return_type(
            class_source, class_names, method_names
        )
        return_class_sources = (
            extract_class_sources(source_code, return_types) if return_types else ""
        )
        return_class_source = "\n".join(return_class_sources)
        rendered_globals = _render_globals(global_vars)
        model_source = "\n".join(
            [
                x
                for x in [imports, rendered_globals, return_class_source, class_source]
                if x
            ]
        )
    else:
        # use class_name specified in cog.yaml as method_name
        method_names = class_names
        function_source = extract_function_source(source_code, method_names)
        if not function_source:
            return None
        function_source = make_function_empty(function_source, method_names)
        if not function_source:
            return None
        return_types = extract_function_return_types(function_source, method_names)
        return_class_sources = (
            extract_class_sources(source_code, return_types) if return_types else ""
        )
        return_class_source = "\n".join(return_class_sources)
        model_source = "\n".join([imports, return_class_source, function_source])
    return model_source



import os

import pytest

from .util import random_string, remove_docker_image


def pytest_sessionstart(session):
    os.environ["COG_NO_UPDATE_CHECK"] = "1"


@pytest.fixture
def docker_image_name():
    return "cog-test-" + random_string(10)


@pytest.fixture
def docker_image(docker_image_name):
    yield docker_image_name
    # We expect the image to exist by this point and will fail if it doesn't.
    # If you just need a name, use docker_image_name.
    remove_docker_image(docker_image_name)



import os.path
import signal
import subprocess
import sys
import time

from cog import BasePredictor


class Predictor(BasePredictor):
    """
    This predictor checks the case where a process is spawned during setup and then each
    prediction depends on being able to communicate with that process. In the event that
    stream redirection is not working correctly, the forked process will not be able to
    write to stdout/stderr and will likely exit. Any state other than "running" is
    considered an error condition and raises SystemExit to interrupt any more prediction
    serving.

    This variant runs a forked python process via a shell wrapper to which a "message" is
    sent via file for each call to `predict`.
    """

    def setup(self) -> None:
        print("---> starting background process")

        self.bg = subprocess.Popen(["bash", "run-forker.sh"])

        print(f"---> started background process pid={self.bg.pid}")

    def predict(self, s: str) -> str:
        status = self.bg.poll()

        print(f"---> background job status={status}")

        if status is not None:
            raise SystemExit

        print(f"---> sending message to background job pid={self.bg.pid}")

        with open(".inbox", "w") as inbox:
            inbox.write(s)

        print(f"---> sent message to background job pid={self.bg.pid}")

        now = time.time()

        print(f"---> waiting for outbox message from background job pid={self.bg.pid}")

        while not os.path.exists(".outbox"):
            if time.time() - now > 5:
                raise TimeoutError

            time.sleep(0.01)

        try:
            with open(".outbox", "r") as outbox:
                print(f"---> relaying message from background job pid={self.bg.pid}")

                return outbox.read()

        finally:
            os.unlink(".outbox")



import base64
import io
import time
import unittest.mock as mock

import pytest
import responses
from PIL import Image
from responses import matchers

from cog.types import PYDANTIC_V2

from .conftest import (
    make_client,
    uses_predictor,
    uses_predictor_with_client_options,
    uses_trainer,
)


def test_index_document():
    client = make_client(fixture_name="slow_setup")
    resp = client.get("/")
    data = resp.json()
    for field in (
        "cog_version",
        "docs_url",
        "openapi_url",
        "shutdown_url",
        "healthcheck_url",
        "predictions_url",
        "predictions_idempotent_url",
        "predictions_cancel_url",
    ):
        assert field in data
        assert data[field] is not None


def test_setup_healthcheck():
    client = make_client(fixture_name="slow_setup")
    resp = client.get("/health-check")
    data = resp.json()
    assert data["status"] == "STARTING"
    assert data["setup"] == {}


@uses_predictor("setup")
def test_setup_is_called(client, match):
    resp = client.post("/predictions")
    assert resp.status_code == 200
    assert resp.json() == match({"status": "succeeded", "output": "bar"})


@uses_predictor("function.py:predict")
def test_predict_works_with_functions(client, match):
    resp = client.post("/predictions", json={"input": {"text": "baz"}})
    assert resp.status_code == 200
    assert resp.json() == match({"status": "succeeded", "output": "hello baz"})


@uses_predictor("openapi_complex_input")
def test_openapi_specification(client):
    resp = client.get("/openapi.json")
    assert resp.status_code == 200

    schema = resp.json()
    assert schema["openapi"] == "3.0.2"
    assert schema["info"] == {"title": "Cog", "version": "0.1.0"}
    assert schema["paths"]["/"] == {
        "get": {
            "summary": "Root",
            "operationId": "root__get",
            "responses": {
                "200": {
                    "description": "Successful Response",
                    "content": {"application/json": {"schema": mock.ANY}},
                }
            },
        }
    }
    assert schema["paths"]["/predictions"] == {
        "post": {
            "summary": "Predict",
            "description": "Run a single prediction on the model",
            "operationId": "predict_predictions_post",
            "parameters": [
                {
                    "in": "header",
                    "name": "prefer",
                    "required": False,
                    "schema": {"title": "Prefer", "type": "string"},
                }
            ],
            "requestBody": {
                "content": {
                    "application/json": {
                        "schema": {"$ref": "#/components/schemas/PredictionRequest"}
                    }
                }
            },
            "responses": {
                "200": {
                    "description": "Successful Response",
                    "content": {
                        "application/json": {
                            "schema": {
                                "$ref": "#/components/schemas/PredictionResponse"
                            }
                        }
                    },
                },
                "422": {
                    "description": "Validation Error",
                    "content": {
                        "application/json": {
                            "schema": {
                                "$ref": "#/components/schemas/HTTPValidationError"
                            }
                        }
                    },
                },
            },
        }
    }
    assert schema["paths"]["/predictions/{prediction_id}/cancel"] == {
        "post": {
            "summary": "Cancel",
            "description": "Cancel a running prediction",
            "operationId": "cancel_predictions__prediction_id__cancel_post",
            "parameters": [
                {
                    "in": "path",
                    "name": "prediction_id",
                    "required": True,
                    "schema": {"title": "Prediction ID", "type": "string"},
                }
            ],
            "responses": {
                "200": {
                    "content": {"application/json": {"schema": mock.ANY}},
                    "description": "Successful Response",
                },
                "422": {
                    "content": {
                        "application/json": {
                            "schema": {
                                "$ref": "#/components/schemas/HTTPValidationError"
                            }
                        }
                    },
                    "description": "Validation Error",
                },
            },
        }
    }
    assert schema["components"]["schemas"]["Input"] == {
        "title": "Input",
        "required": [
            "no_default",
            "path",
            "image",
            "choices",
            "int_choices",
        ],
        "type": "object",
        "properties": {
            "no_default": {
                "title": "No Default",
                "type": "string",
                "x-order": 0,
            },
            "default_without_input": {
                "title": "Default Without Input",
                "type": "string",
                "default": "default",
                "x-order": 1,
            },
            "input_with_default": {
                "title": "Input With Default",
                "type": "integer",
                "default": -10,
                "x-order": 2,
            },
            "path": {
                "title": "Path",
                "description": "Some path",
                "type": "string",
                "format": "uri",
                "x-order": 3,
            },
            "image": {
                "title": "Image",
                "description": "Some path",
                "type": "string",
                "format": "uri",
                "x-order": 4,
            },
            "choices": {
                "allOf": [{"$ref": "#/components/schemas/choices"}],
                "x-order": 5,
            },
            "int_choices": {
                "allOf": [{"$ref": "#/components/schemas/int_choices"}],
                "x-order": 6,
            },
        },
    }
    assert schema["components"]["schemas"]["Output"] == {
        "title": "Output",
        "type": "string",
    }
    assert schema["components"]["schemas"]["choices"] == {
        "title": "choices",
        "enum": ["foo", "bar"],
        "description": "An enumeration.",
        "type": "string",
    }
    assert schema["components"]["schemas"]["int_choices"] == {
        "description": "An enumeration.",
        "enum": [3, 4, 5],
        "title": "int_choices",
        "type": "integer",
    }


@uses_predictor("openapi_custom_output_type")
def test_openapi_specification_with_custom_user_defined_output_type(
    client,
):
    resp = client.get("/openapi.json")
    assert resp.status_code == 200

    schema = resp.json()
    assert schema["components"]["schemas"]["Output"] == {
        "$ref": "#/components/schemas/MyOutput",
        "title": "Output",
    }
    assert schema["components"]["schemas"]["MyOutput"] == {
        "title": "MyOutput",
        "type": "object",
        "properties": {
            "foo_number": {
                "title": "Foo Number",
                "type": "integer",
                "default": "42",
            },
            "foo_string": {
                "title": "Foo String",
                "type": "string",
                "default": "meaning of life",
            },
        },
    }


@uses_predictor("openapi_output_type")
def test_openapi_specification_with_custom_user_defined_output_type_called_output(
    client,
):
    resp = client.get("/openapi.json")
    assert resp.status_code == 200
    schema = resp.json()
    assert resp.json()["components"]["schemas"]["Output"] == {
        "properties": {
            "foo_number": {"default": "42", "title": "Foo Number", "type": "integer"},
            "foo_string": {
                "default": "meaning of life",
                "title": "Foo String",
                "type": "string",
            },
        },
        "title": "Output",
        "type": "object",
    }


@uses_predictor("openapi_output_yield")
def test_openapi_specification_with_yield(client):
    resp = client.get("/openapi.json")
    assert resp.status_code == 200
    schema = resp.json()
    assert schema["components"]["schemas"]["Output"] == {
        "title": "Output",
        "type": "array",
        "items": {
            "type": "string",
        },
        "x-cog-array-type": "iterator",
    }


@uses_predictor("yield_concatenate_iterator")
def test_openapi_specification_with_yield_with_concatenate_iterator(client):
    resp = client.get("/openapi.json")
    assert resp.status_code == 200

    schema = resp.json()
    assert schema["components"]["schemas"]["Output"] == {
        "title": "Output",
        "type": "array",
        "items": {
            "type": "string",
        },
        "x-cog-array-type": "iterator",
        "x-cog-array-display": "concatenate",
    }


@uses_predictor("openapi_output_list")
def test_openapi_specification_with_list(client):
    resp = client.get("/openapi.json")
    assert resp.status_code == 200

    schema = resp.json()
    assert schema["components"]["schemas"]["Output"] == {
        "title": "Output",
        "type": "array",
        "items": {
            "type": "string",
        },
    }


@uses_predictor("openapi_input_int_choices")
def test_openapi_specification_with_int_choices(client):
    resp = client.get("/openapi.json")
    assert resp.status_code == 200

    schema = resp.json()
    schemas = schema["components"]["schemas"]

    assert schemas["Input"]["properties"]["pick_a_number_any_number"] == {
        "allOf": [{"$ref": "#/components/schemas/pick_a_number_any_number"}],
        "x-order": 0,
    }
    assert schemas["pick_a_number_any_number"] == {
        "description": "An enumeration.",
        "enum": [1, 2],
        "title": "pick_a_number_any_number",
        "type": "integer",
    }


@uses_trainer("train.py:train")
def test_train_openapi_specification(client):
    resp = client.get("/openapi.json")
    assert resp.status_code == 200

    schema = resp.json()
    assert schema["openapi"] == "3.0.2"
    assert schema["info"] == {"title": "Cog", "version": "0.1.0"}

    assert schema["components"]["schemas"]["TrainingInput"] == {
        "title": "TrainingInput",
        "type": "object",
        "properties": {
            "n": {
                "type": "integer",
                "x-order": 0,
                "title": "N",
                "description": "Dimension of weights to generate",
            },
        },
        "required": [
            "n",
        ],
    }

    assert schema["components"]["schemas"]["TrainingOutput"] == {
        "title": "TrainingOutput",
        "type": "object",
        "properties": {
            "weights": {
                "type": "string",
                "format": "uri",
                "title": "Weights",
            },
        },
        "required": [
            "weights",
        ],
    }


@pytest.mark.skipif(
    not PYDANTIC_V2,
    reason="Literal is used for enums only in Pydantic v2",
)
@uses_predictor("input_literal")
def test_openapi_specification_with_literal(client, static_schema):
    resp = client.get("/openapi.json")
    assert resp.status_code == 200

    schema = resp.json()
    assert schema["openapi"] == "3.0.2"
    assert schema["info"] == {"title": "Cog", "version": "0.1.0"}

    schemas = schema["components"]["schemas"]

    assert schemas["Input"]["properties"]["text"] == {
        "allOf": [{"$ref": "#/components/schemas/text"}],
        "x-order": 0,
    }
    assert schemas["text"] == {
        "description": "An enumeration.",
        "enum": ["foo", "bar"],
        "title": "text",
        "type": "string",
    }


@uses_predictor("yield_strings")
def test_yielding_strings_from_generator_predictors(client, match):
    resp = client.post("/predictions")
    assert resp.status_code == 200
    assert resp.json() == match(
        {"status": "succeeded", "output": ["foo", "bar", "baz"]}
    )


@uses_predictor("yield_concatenate_iterator")
def test_yielding_strings_from_concatenate_iterator(client, match):
    resp = client.post("/predictions")
    assert resp.status_code == 200
    assert resp.json() == match(
        {"status": "succeeded", "output": ["foo", "bar", "baz"]}
    )


@uses_predictor("yield_strings_file_input")
def test_yielding_strings_from_generator_predictors_file_input(client, match):
    resp = client.post(
        "/predictions",
        json={"input": {"file": "data:text/plain; charset=utf-8;base64,aGVsbG8="}},
    )
    assert resp.status_code == 200
    assert resp.json() == match(
        {
            "status": "succeeded",
            "output": ["hello foo", "hello bar", "hello baz"],
        }
    )


@uses_predictor("yield_files")
def test_yielding_files_from_generator_predictors(client):
    resp = client.post("/predictions")

    assert resp.status_code == 200
    output = resp.json()["output"]

    def image_color(data_url):
        header, b64data = data_url.split(",", 1)
        image = Image.open(io.BytesIO(base64.b64decode(b64data)))
        return Image.Image.getcolors(image)[0][1]

    assert image_color(output[0]) == (255, 0, 0)  # red
    assert image_color(output[1]) == (0, 0, 255)  # blue
    assert image_color(output[2]) == (255, 255, 0)  # yellow


@uses_predictor("input_none")
def test_prediction_idempotent_endpoint(client, match):
    resp = client.put("/predictions/abcd1234", json={})
    assert resp.status_code == 200
    assert resp.json() == match(
        {"id": "abcd1234", "status": "succeeded", "output": "foobar"}
    )


@uses_predictor("input_none")
def test_prediction_idempotent_endpoint_matched_ids(client, match):
    resp = client.put(
        "/predictions/abcd1234",
        json={
            "id": "abcd1234",
        },
    )
    assert resp.status_code == 200
    assert resp.json() == match(
        {"id": "abcd1234", "status": "succeeded", "output": "foobar"}
    )


@uses_predictor("input_none")
def test_prediction_idempotent_endpoint_mismatched_ids(client, match):
    resp = client.put(
        "/predictions/abcd1234",
        json={
            "id": "foobar",
        },
    )
    assert resp.status_code == 422


@uses_predictor("sleep")
def test_prediction_idempotent_endpoint_is_idempotent(client, match):
    resp1 = client.put(
        "/predictions/abcd1234",
        json={"input": {"sleep": 1}},
        headers={"Prefer": "respond-async"},
    )
    resp2 = client.put(
        "/predictions/abcd1234",
        json={"input": {"sleep": 1}},
        headers={"Prefer": "respond-async"},
    )
    assert resp1.status_code == 202
    assert resp1.json() == match({"id": "abcd1234", "status": "processing"})
    assert resp2.status_code == 202
    assert resp2.json() == match({"id": "abcd1234", "status": "processing"})


@uses_predictor("sleep")
def test_prediction_idempotent_endpoint_conflict(client, match):
    resp1 = client.put(
        "/predictions/abcd1234",
        json={"input": {"sleep": 1}},
        headers={"Prefer": "respond-async"},
    )
    resp2 = client.put(
        "/predictions/5678efgh",
        json={"input": {"sleep": 1}},
        headers={"Prefer": "respond-async"},
    )
    assert resp1.status_code == 202
    assert resp1.json() == match({"id": "abcd1234", "status": "processing"})
    assert resp2.status_code == 409


# a basic end-to-end test for async predictions. if you're adding more
# exhaustive tests of webhooks, consider adding them to test_runner.py
@responses.activate
@uses_predictor("input_string")
def test_asynchronous_prediction_endpoint(client, match):
    webhook = responses.post(
        "https://example.com/webhook",
        match=[
            matchers.json_params_matcher(
                {
                    "id": "12345abcde",
                    "status": "succeeded",
                    "output": "hello world",
                },
                strict_match=False,
            )
        ],
        status=200,
    )

    resp = client.post(
        "/predictions",
        json={
            "id": "12345abcde",
            "input": {"text": "hello world"},
            "webhook": "https://example.com/webhook",
            "webhook_events_filter": ["completed"],
        },
        headers={"Prefer": "respond-async"},
    )
    assert resp.status_code == 202

    result = resp.json()

    # The response might be a "processing" response, but the prediction can
    # also complete before the response is sent.
    assert result["started_at"] is not None
    assert result["status"] in {"processing", "succeeded"}
    assert result["output"] in {None, "hello world"}

    n = 0
    while webhook.call_count < 1 and n < 10:
        time.sleep(0.1)
        n += 1


# End-to-end test for passing tracing headers on to downstream services.
@responses.activate
@uses_predictor_with_client_options(
    "output_file", upload_url="https://example.com/upload"
)
def test_asynchronous_prediction_endpoint_with_trace_context(client, match):
    webhook = responses.post(
        "https://example.com/webhook",
        match=[
            matchers.json_params_matcher(
                {
                    "id": "12345abcde",
                    "status": "succeeded",
                    "output": "https://example.com/upload/file",
                },
                strict_match=False,
            ),
            matchers.header_matcher(
                {
                    "traceparent": "traceparent-123",
                    "tracestate": "tracestate-123",
                },
                strict_match=False,
            ),
        ],
        status=200,
    )
    uploader = responses.put(
        "https://example.com/upload/file",
        match=[
            matchers.header_matcher(
                {
                    "traceparent": "traceparent-123",
                    "tracestate": "tracestate-123",
                },
                strict_match=False,
            ),
        ],
        status=200,
    )

    resp = client.post(
        "/predictions",
        json={
            "id": "12345abcde",
            "input": {},
            "webhook": "https://example.com/webhook",
            "webhook_events_filter": ["completed"],
        },
        headers={
            "Prefer": "respond-async",
            "traceparent": "traceparent-123",
            "tracestate": "tracestate-123",
        },
    )
    assert resp.status_code == 202

    result = resp.json()

    # The response might be a "processing" response, but the prediction can
    # also complete before the response is sent.
    assert result["started_at"] is not None
    assert result["status"] in {"processing", "succeeded"}
    assert result["output"] in {None, "https://example.com/upload/file"}

    n = 0
    while webhook.call_count < 1 and n < 10:
        time.sleep(0.1)
        n += 1

    assert uploader.call_count == 1
    assert webhook.call_count == 1


@uses_predictor("sleep")
def test_prediction_cancel(client):
    resp = client.post("/predictions/123/cancel")
    assert resp.status_code == 404

    resp = client.post(
        "/predictions",
        json={"id": "123", "input": {"sleep": 1}},
        headers={"Prefer": "respond-async"},
    )
    assert resp.status_code == 202

    resp = client.post("/predictions/456/cancel")
    assert resp.status_code == 404

    resp = client.post("/predictions/123/cancel")
    assert resp.status_code == 200


@uses_predictor_with_client_options(
    "setup_weights",
    env={"COG_WEIGHTS": "data:text/plain; charset=utf-8;base64,aGVsbG8="},
)
def test_weights_are_read_from_environment_variables(client, match):
    resp = client.post("/predictions")
    assert resp.status_code == 200
    assert resp.json() == match({"status": "succeeded", "output": "hello"})



from pathlib import Path

import pydantic
from pydantic import BaseModel

from .types import PYDANTIC_V2, URLPath


# Base class for inputs, constructed dynamically in get_input_type().
# (This can't be a docstring or it gets passed through to the schema.)
class BaseInput(BaseModel):
    if PYDANTIC_V2:
        model_config = pydantic.ConfigDict(use_enum_values=True)  # type: ignore
    else:

        class Config:
            # When using `choices`, the type is converted into an enum to validate
            # But, after validation, we want to pass the actual value to predict(), not the enum object
            use_enum_values = True

    def cleanup(self) -> None:
        """
        Cleanup any temporary files created by the input.
        """
        for _, value in self:
            # Handle URLPath objects specially for cleanup.
            # Also handle pathlib.Path objects, which cog.Path is a subclass of.
            # A pathlib.Path object shouldn't make its way here,
            # but both have an unlink() method, so we may as well be safe.
            if isinstance(value, (URLPath, Path)):
                value.unlink(missing_ok=True)



import time


class Predictor:
    def setup(self):
        print("did setup")

    def predict(self):
        for _ in range(10):
            print("doing stuff")
            time.sleep(3)
        print("did predict")



from cog import BasePredictor
from pydantic import BaseModel


# An output object called `Output` needs to be special cased because pydantic tries to dedupe it with the internal `Output`
class Output(BaseModel):
    foo_number: int = "42"
    foo_string: str = "meaning of life"


class Predictor(BasePredictor):
    def predict(
        self,
    ) -> Output:
        pass



import asyncio
import contextlib
import inspect
import multiprocessing
import os
import signal
import sys
import threading
import traceback
import types
import uuid
import weakref
from concurrent import futures
from concurrent.futures import Future, ThreadPoolExecutor
from enum import Enum, auto, unique
from multiprocessing.connection import Connection
from typing import (
    Any,
    Callable,
    Dict,
    Iterator,
    Optional,
    Tuple,
    Union,
    cast,
)

import structlog
from attrs import define

from ..base_predictor import BasePredictor
from ..json import make_encodeable
from ..predictor import (
    extract_setup_weights,
    get_predict,
    has_setup_weights,
    load_predictor_from_ref,
)
from ..types import PYDANTIC_V2, URLPath
from ..wait import wait_for_env
from .connection import AsyncConnection, LockedConnection
from .eventtypes import (
    Cancel,
    Done,
    Envelope,
    Log,
    PredictionInput,
    PredictionMetric,
    PredictionOutput,
    PredictionOutputType,
    Shutdown,
)
from .exceptions import (
    CancelationException,
    FatalWorkerException,
    InvalidStateException,
)
from .helpers import SimpleStreamRedirector, StreamRedirector
from .scope import Scope, _get_current_scope, evolve_scope, scope

if PYDANTIC_V2:
    from .helpers import unwrap_pydantic_serialization_iterators

_spawn = multiprocessing.get_context("spawn")

_PublicEventType = Union[Done, Log, PredictionOutput, PredictionOutputType]

log = structlog.get_logger("cog.server.worker")


@unique
class WorkerState(Enum):
    NEW = auto()
    STARTING = auto()
    READY = auto()
    DEFUNCT = auto()


@define
class PredictionRequest:
    tag: Optional[str]


@define
class CancelRequest:
    tag: Optional[str]


@define
class PredictionState:
    tag: Optional[str]
    payload: Dict[str, Any]
    result: "Future[Done]"

    cancel_sent: bool = False


class Worker:
    @property
    def uses_concurrency(self) -> bool:
        return self._max_concurrency > 1

    def __init__(
        self, child: "_ChildWorker", events: Connection, max_concurrency: int = 1
    ) -> None:
        self._child = child
        self._events = events

        self._sent_shutdown_event = False
        self._state = WorkerState.NEW
        self._terminating = False

        self._setup_result: "Future[Done]" = Future()
        self._subscribers_lock = threading.Lock()
        self._subscribers: Dict[
            int, Tuple[Callable[[_PublicEventType], None], Optional[str]]
        ] = {}

        self._max_concurrency = max_concurrency

        self._predictions_lock = threading.Lock()
        self._predictions_in_flight: Dict[Optional[str], PredictionState] = {}

        self._event_consumer_pool = ThreadPoolExecutor(max_workers=1)
        self._prediction_start_pool = ThreadPoolExecutor(max_workers=max_concurrency)
        self._input_download_pool = ThreadPoolExecutor(max_workers=8)
        self._event_consumer = None

    def setup(self) -> "Future[Done]":
        self._assert_state(WorkerState.NEW)
        self._state = WorkerState.STARTING
        self._child.start()
        self._event_consumer = self._event_consumer_pool.submit(self._consume_events)
        return self._setup_result

    def predict(
        self, payload: Dict[str, Any], tag: Optional[str] = None
    ) -> "Future[Done]":
        # TODO: tag is Optional, but it's required when in concurrent mode and
        # basically unnecessary in sequential mode. Should we have a separate
        # ConcurrentWorker?
        if self._max_concurrency > 1 and tag is None:
            raise TypeError(
                "Invalid operation: tag is required when Worker has max_concurrency > 1"
            )

        with self._predictions_lock:
            if len(self._predictions_in_flight) >= self._max_concurrency:
                raise InvalidStateException(
                    "Invalid operation: maximum predictions in flight reached"
                )
            if tag in self._predictions_in_flight:
                raise InvalidStateException(
                    f"Invalid operation: prediction with tag {tag} already running"
                )
            self._assert_state(WorkerState.READY)
            result = Future()
            self._predictions_in_flight[tag] = PredictionState(tag, payload, result)

        self._prediction_start_pool.submit(self._start_prediction(tag, payload))
        return result

    def _start_prediction(
        self, tag: Optional[str], payload: Dict[str, Any]
    ) -> Callable[[], None]:
        def start_prediction() -> None:
            try:
                to_await = []
                futs = {}
                # Prepare payload asynchronously (download URLPath objects)
                for k, v in payload.items():
                    # Check if v is an instance of URLPath
                    if isinstance(v, URLPath):
                        futs[k] = self._input_download_pool.submit(v.convert)
                        to_await.append(futs[k])
                    # Check if v is a list of URLPath instances
                    elif isinstance(v, list) and all(
                        isinstance(item, URLPath) for item in v
                    ):
                        futs[k] = [
                            self._input_download_pool.submit(item.convert) for item in v
                        ]
                        to_await += futs[k]
                done, not_done = futures.wait(
                    to_await, return_when=futures.FIRST_EXCEPTION
                )

                if len(not_done) > 0:
                    # if any future isn't done, this is because one of the
                    # futures raised an exception. first we cancel outstanding
                    # work
                    for fut in not_done:
                        fut.cancel()
                    # then we find an exception to raise
                    for fut in done:
                        fut.result()  # raises if the future finished with an exception
                    # we should never get here
                    raise Exception(
                        "Internal error: lost track of exception while downloading input files"
                    )

                # all futures are done. some might still have raised an
                # exception, but when we call fut.result() that will re-raise
                # and do the right thing
                for k, v in futs.items():
                    if isinstance(v, list):
                        payload[k] = []
                        for fut in v:
                            payload[k].append(fut.result())
                    elif isinstance(v, Future):
                        payload[k] = v.result()
                # send the prediction to the child to start
                self._events.send(
                    Envelope(
                        event=PredictionInput(payload=payload),
                        tag=tag,
                    )
                )
            except Exception as e:
                done = Done(error=True, error_detail=str(e))
                self._publish(Envelope(done, tag))
                self._complete_prediction(done, tag)

        return start_prediction

    def subscribe(
        self,
        subscriber: Callable[[_PublicEventType], None],
        tag: Optional[str] = None,
    ) -> int:
        sid = uuid.uuid4().int
        with self._subscribers_lock:
            self._subscribers[sid] = (subscriber, tag)
        return sid

    def unsubscribe(self, sid: int) -> None:
        with self._subscribers_lock:
            del self._subscribers[sid]

    def shutdown(self, timeout: Optional[float] = None) -> None:
        """
        Shut down the worker gracefully. This waits for the child worker to
        finish any in-flight work and exit.
        """
        self._terminating = True
        self._state = WorkerState.DEFUNCT

        if self._child.is_alive() and not self._sent_shutdown_event:
            self._events.send(Envelope(event=Shutdown()))
            self._sent_shutdown_event = True

        if self._event_consumer:
            self._event_consumer.result(timeout=timeout)

        self._event_consumer_pool.shutdown()

    def terminate(self) -> None:
        """
        Shut down the worker immediately. This may not correctly clean up
        resources used by the worker.
        """
        self._terminating = True
        self._state = WorkerState.DEFUNCT

        if self._child.is_alive():
            self._child.terminate()
            self._child.join()

        self._event_consumer_pool.shutdown(wait=False)

    def cancel(self, tag: Optional[str] = None) -> None:
        with self._predictions_lock:
            predict_state = self._predictions_in_flight.get(tag)
            if predict_state and not predict_state.cancel_sent:
                self._child.send_cancel_signal()
                self._events.send(Envelope(event=Cancel(), tag=tag))
                predict_state.cancel_sent = True

    def _assert_state(self, state: WorkerState) -> None:
        if self._state != state:
            raise InvalidStateException(
                f"Invalid operation: state is {self._state} (must be {state})"
            )

    def _consume_events_until_done(self) -> Optional[Done]:
        while self._child.is_alive():
            if not self._events.poll(0.1):
                continue

            e = self._events.recv()
            self._publish(e)

            if isinstance(e.event, Done):
                return e.event
        return None

    def _consume_events(self) -> None:
        try:
            self._consume_events_inner()
        except:
            log.fatal("unhandled error in _consume_events", exc_info=True)
            raise

    def _consume_events_inner(self) -> None:
        # Setup
        done = self._consume_events_until_done()
        # If we didn't get a done event, the child process died.
        if not done:
            exitcode = self._child.exitcode
            self._setup_result.set_exception(
                FatalWorkerException(
                    f"Predictor setup failed for an unknown reason. It might have run out of memory? (exitcode {exitcode})"
                )
            )
            self._state = WorkerState.DEFUNCT
            return
        if done.error:
            self._setup_result.set_exception(
                FatalWorkerException(
                    "Predictor errored during setup: " + done.error_detail
                )
            )
            self._state = WorkerState.DEFUNCT
            return

        # We capture the setup future and then set state to READY before
        # completing it, so that we can immediately accept work.
        self._state = WorkerState.READY
        self._setup_result.set_result(done)

        # Main event loop
        while self._child.is_alive():
            # wait for events from the child worker
            if not self._events.poll(0.1):
                continue

            ev = self._events.recv()
            self._publish(ev)
            if isinstance(ev.event, Done):
                self._complete_prediction(ev.event, ev.tag)

        # If we dropped off the end off the end of the loop, it's because the
        # child process died.  First, process any remaining messages on the connection
        while self._events.poll():
            ev = self._events.recv()
            self._publish(ev)
            if isinstance(ev.event, Done):
                self._complete_prediction(ev.event, ev.tag)

        if not self._terminating:
            self._state = WorkerState.DEFUNCT
            with self._predictions_lock:
                for state in self._predictions_in_flight.values():
                    exitcode = self._child.exitcode
                    state.result.set_exception(
                        FatalWorkerException(
                            f"Prediction failed for an unknown reason. It might have run out of memory? (exitcode {exitcode})"
                        )
                    )
                self._predictions_in_flight.clear()

    def _complete_prediction(self, done: Done, tag: Optional[str]) -> None:
        # We update the in-flight dictionary before completing the prediction
        # future, so that we can immediately accept work.
        with self._predictions_lock:
            predict_state = self._predictions_in_flight.pop(tag)
        predict_state.result.set_result(done)

    def _publish(self, e: Envelope) -> None:
        with self._subscribers_lock:
            subscribers_copy = list(self._subscribers.values())
        for subscriber, requested_tag in subscribers_copy:
            if requested_tag is None or e.tag == requested_tag:
                try:
                    subscriber(cast(_PublicEventType, e.event))
                except Exception:
                    log.warn(
                        "publish failed",
                        subscriber=subscriber,
                        tag=e.tag,
                        event=e.event,
                        exc_info=True,
                    )


class _ChildWorker(_spawn.Process):  # type: ignore
    def __init__(
        self,
        predictor_ref: str,
        *,
        is_async: bool,
        events: Connection,
        max_concurrency: int = 1,
        tee_output: bool = True,
    ) -> None:
        self._predictor_ref = predictor_ref
        self._predictor: Optional[BasePredictor] = None
        self._events: Union[AsyncConnection, LockedConnection] = LockedConnection(
            events
        )
        self._tee_output = tee_output
        self._cancelable = False
        self._max_concurrency = max_concurrency

        # for synchronous predictors only! async predictors use current_scope()._tag instead
        self._sync_tag: Optional[str] = None
        self._has_async_predictor = is_async

        super().__init__()

    def run(self) -> None:
        # If we're running at a shell, SIGINT will be sent to every process in
        # the process group. We ignore it in the child process and require that
        # shutdown is coordinated by the parent process.
        signal.signal(signal.SIGINT, signal.SIG_IGN)

        # Initially, we ignore SIGUSR1.
        signal.signal(signal.SIGUSR1, signal.SIG_IGN)

        if self._has_async_predictor:
            redirector = SimpleStreamRedirector(
                callback=self._stream_write_hook,
                tee=self._tee_output,
            )
        else:
            redirector = StreamRedirector(
                callback=self._stream_write_hook,
                tee=self._tee_output,
            )

        with scope(Scope(record_metric=self.record_metric)), redirector:
            with self._handle_setup_error(redirector):
                wait_for_env()
                self._predictor = load_predictor_from_ref(self._predictor_ref)

            # If load_predictor_from_ref hasn't returned a predictor instance then
            # it has sent a error Done event and we're done here.
            if not self._predictor:
                return

            if not self._validate_predictor(redirector):
                return

            predict = get_predict(self._predictor)

            if self._has_async_predictor:
                assert isinstance(redirector, SimpleStreamRedirector)
                predictor = self._predictor

                async def _runner() -> None:
                    if hasattr(predictor, "setup") and inspect.iscoroutinefunction(
                        predictor.setup
                    ):
                        await self._asetup(redirector)
                    else:
                        self._setup(redirector)
                    await self._aloop(predict, redirector)

                asyncio.run(_runner())
            else:
                # We use SIGUSR1 to signal an interrupt for cancelation.
                signal.signal(signal.SIGUSR1, self._signal_handler)

                assert isinstance(redirector, StreamRedirector)
                self._setup(redirector)
                self._loop(
                    predict,
                    redirector,
                )

    def send_cancel_signal(self) -> None:
        if self.is_alive() and self.pid:
            os.kill(self.pid, signal.SIGUSR1)

    def record_metric(self, name: str, value: Union[float, int]) -> None:
        self._events.send(
            Envelope(PredictionMetric(name, value), tag=self._current_tag)
        )

    @property
    def _current_tag(self) -> Optional[str]:
        if self._has_async_predictor:
            return _get_current_scope()._tag
        return self._sync_tag

    def _validate_predictor(
        self,
        redirector: Union[StreamRedirector, SimpleStreamRedirector],
    ) -> bool:
        with self._handle_setup_error(redirector):
            assert self._predictor

            # Async models require python >= 3.11 so we can use asyncio.TaskGroup
            # We should check for this before getting to this point
            if self._has_async_predictor and sys.version_info < (3, 11):
                raise FatalWorkerException(
                    "Cog requires Python >=3.11 for `async def predict()` support"
                )

            if self._max_concurrency > 1 and not self._has_async_predictor:
                raise FatalWorkerException(
                    "max_concurrency > 1 requires an async predict function, e.g. `async def predict()`"
                )

            if (
                hasattr(self._predictor, "setup")
                and inspect.iscoroutinefunction(self._predictor.setup)
                and not self._has_async_predictor
            ):
                raise FatalWorkerException(
                    "Invalid predictor: to use an async setup method you must use an async predict method"
                )

            return True

        return False

    def _setup(
        self, redirector: Union[StreamRedirector, SimpleStreamRedirector]
    ) -> None:
        with self._handle_setup_error(redirector, ensure_done_event=True):
            assert self._predictor

            # Could be a function or a class
            if not hasattr(self._predictor, "setup"):
                return

            if not has_setup_weights(self._predictor):
                self._predictor.setup()
                return

            weights = extract_setup_weights(self._predictor)
            self._predictor.setup(weights=weights)  # type: ignore

    async def _asetup(
        self, redirector: Union[StreamRedirector, SimpleStreamRedirector]
    ) -> None:
        with self._handle_setup_error(redirector, ensure_done_event=True):
            assert self._predictor

            # Could be a function or a class
            if not hasattr(self._predictor, "setup"):
                return

            if not has_setup_weights(self._predictor):
                await self._predictor.setup()  # type: ignore
                return

            weights = extract_setup_weights(self._predictor)
            await self._predictor.setup(weights=weights)  # type: ignore

    def _loop(
        self,
        predict: Callable[..., Any],
        redirector: StreamRedirector,
    ) -> None:
        while True:
            e = cast(Envelope, self._events.recv())
            if isinstance(e.event, Cancel):
                # for sync predictors, this is handled via SIGUSR1 signals from
                # the parent via send_cancel_signal
                continue
            elif isinstance(e.event, Shutdown):
                break
            elif isinstance(e.event, PredictionInput):
                self._predict(e.tag, e.event.payload, predict, redirector)
            else:
                print(f"Got unexpected event: {e.event}", file=sys.stderr)

    async def _aloop(
        self,
        predict: Callable[..., Any],
        redirector: SimpleStreamRedirector,
    ) -> None:
        # Unwrap and replace the events connection with an async one.
        assert isinstance(self._events, LockedConnection)
        self._events = AsyncConnection(self._events.connection)

        async with asyncio.TaskGroup() as tg:
            tasks = weakref.WeakValueDictionary[str | None, asyncio.Task[Any]]()
            while True:
                e = cast(Envelope, await self._events.recv())
                if isinstance(e.event, Cancel):
                    # NOTE: We don't check the _cancelable flag here, instead we rely
                    # on the presence of the value in the weakmap to determine if
                    # a prediction is actively being processed.
                    task = tasks.get(e.tag)
                    if not task:
                        print(
                            "Got cancel event for unrecognized prediction",
                            file=sys.stderr,
                        )
                        continue

                    task.cancel()
                elif isinstance(e.event, Shutdown):
                    break
                elif isinstance(e.event, PredictionInput):
                    tasks[e.tag] = tg.create_task(
                        self._apredict(e.tag, e.event.payload, predict, redirector)
                    )
                else:
                    print(f"Got unexpected event: {e.event}", file=sys.stderr)

    def _predict(
        self,
        tag: Optional[str],
        payload: Dict[str, Any],
        predict: Callable[..., Any],
        redirector: StreamRedirector,
    ) -> None:
        with self._handle_predict_error(redirector, tag=tag):
            result = predict(**payload)

            if result:
                if isinstance(result, types.GeneratorType):
                    self._events.send(
                        Envelope(
                            event=PredictionOutputType(multi=True),
                            tag=tag,
                        )
                    )
                    for r in result:
                        if PYDANTIC_V2:
                            payload = make_encodeable(
                                unwrap_pydantic_serialization_iterators(r)
                            )
                        else:
                            payload = make_encodeable(r)
                        self._events.send(
                            Envelope(
                                event=PredictionOutput(payload=payload),
                                tag=tag,
                            )
                        )
                else:
                    self._events.send(
                        Envelope(
                            event=PredictionOutputType(multi=False),
                            tag=tag,
                        )
                    )
                    if PYDANTIC_V2:
                        payload = make_encodeable(
                            unwrap_pydantic_serialization_iterators(result)
                        )
                    else:
                        payload = make_encodeable(result)
                    self._events.send(
                        Envelope(
                            event=PredictionOutput(payload=payload),
                            tag=tag,
                        )
                    )

    async def _apredict(
        self,
        tag: Optional[str],
        payload: Dict[str, Any],
        predict: Callable[..., Any],
        redirector: SimpleStreamRedirector,
    ) -> None:
        with evolve_scope(tag=tag), self._handle_predict_error(redirector, tag=tag):
            future_result = predict(**payload)

            if future_result:
                if inspect.isasyncgen(future_result):
                    self._events.send(
                        Envelope(
                            event=PredictionOutputType(multi=True),
                            tag=tag,
                        )
                    )
                    async for r in future_result:
                        if PYDANTIC_V2:
                            payload = make_encodeable(
                                unwrap_pydantic_serialization_iterators(r)
                            )
                        else:
                            payload = make_encodeable(r)
                        self._events.send(
                            Envelope(
                                event=PredictionOutput(payload=payload),
                                tag=tag,
                            )
                        )
                else:
                    result = await future_result
                    self._events.send(
                        Envelope(
                            event=PredictionOutputType(multi=False),
                            tag=tag,
                        )
                    )
                    if PYDANTIC_V2:
                        payload = make_encodeable(
                            unwrap_pydantic_serialization_iterators(result)
                        )
                    else:
                        payload = make_encodeable(result)
                    self._events.send(
                        Envelope(
                            event=PredictionOutput(payload=payload),
                            tag=tag,
                        )
                    )

    @contextlib.contextmanager
    def _handle_setup_error(
        self,
        redirector: Union[SimpleStreamRedirector, StreamRedirector],
        *,
        ensure_done_event: bool = False,
    ) -> Iterator[None]:
        done = Done()
        try:
            yield
        except Exception as e:  # pylint: disable=broad-exception-caught
            traceback.print_exc()
            done.error = True
            done.error_detail = str(e)
        except BaseException as e:
            # For SystemExit and friends we attempt to add some useful context
            # to the logs, but reraise to ensure the process dies.
            traceback.print_exc()
            done.error = True
            done.error_detail = str(e)
            raise
        finally:
            try:
                redirector.drain(timeout=10)
            except TimeoutError:
                self._events.send(
                    Envelope(
                        event=Log(
                            "WARNING: logs may be truncated due to excessive volume.",
                            source="stderr",
                        )
                    )
                )
                raise

            if done.error or ensure_done_event:
                self._events.send(Envelope(event=done))

    @contextlib.contextmanager
    def _handle_predict_error(
        self,
        redirector: Union[SimpleStreamRedirector, StreamRedirector],
        tag: Optional[str],
    ) -> Iterator[None]:
        done = Done()
        send_done = True
        self._cancelable = True
        self._sync_tag = tag
        try:
            yield
        # regular cancelation
        except CancelationException:
            done.canceled = True
        # async cancelation
        except asyncio.CancelledError:
            done.canceled = True
            # We've handled the requested cancelation, so we uncancel the task.
            # This ensures that any cleanup work we do won't be interrupted.
            task = asyncio.current_task()
            assert task
            task.uncancel()
        except Exception as e:  # pylint: disable=broad-exception-caught
            traceback.print_exc()
            done.error = True
            done.error_detail = str(e)
        except BaseException:
            # For SystemExit and friends we attempt to add some useful context
            # to the logs, but reraise to ensure the process dies.
            traceback.print_exc()
            # This is fatal, so we should not send a done event, as this
            # implies we're ready for more work.
            send_done = False
            raise
        finally:
            self._cancelable = False
            try:
                redirector.drain(timeout=10)
            except TimeoutError:
                self._events.send(
                    Envelope(
                        event=Log(
                            "WARNING: logs may be truncated due to excessive volume.",
                            source="stderr",
                        ),
                        tag=tag,
                    )
                )
                raise
            if send_done:
                self._events.send(Envelope(event=done, tag=tag))
            self._sync_tag = None

    def _signal_handler(
        self,
        signum: int,
        frame: Optional[types.FrameType],  # pylint: disable=unused-argument
    ) -> None:
        if signum == signal.SIGUSR1 and self._cancelable:
            raise CancelationException()

    def _stream_write_hook(self, stream_name: str, data: str) -> None:
        if len(data) == 0:
            return

        if stream_name == sys.stdout.name:
            self._events.send(
                Envelope(event=Log(data, source="stdout"), tag=self._current_tag)
            )
        else:
            self._events.send(
                Envelope(event=Log(data, source="stderr"), tag=self._current_tag)
            )


def make_worker(
    predictor_ref: str,
    *,
    is_async: bool,
    tee_output: bool = True,
    max_concurrency: int = 1,
) -> Worker:
    parent_conn, child_conn = _spawn.Pipe()
    child = _ChildWorker(
        predictor_ref,
        is_async=is_async,
        events=child_conn,
        tee_output=tee_output,
        max_concurrency=max_concurrency,
    )
    parent = Worker(child=child, events=parent_conn, max_concurrency=max_concurrency)
    return parent



from cog import BasePredictor, Input


class Predictor(BasePredictor):
    def predict(self, text: str = Input(choices={"foo": "x", "bar": "y"}.keys())) -> str:
        assert type(text) == str
        return text
